# Copyright 2018 Intel, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import fixtures
import os
from unittest import mock

from cyborg.accelerator.drivers.fpga.intel.driver import IntelFPGADriver
from cyborg.accelerator.drivers.fpga.intel import sysinfo
from cyborg.tests import base
from cyborg.tests.unit.accelerator.drivers.fpga.intel import prepare_test_data


class TestIntelFPGADriver(base.TestCase):

    def setUp(self):
        super(TestIntelFPGADriver, self).setUp()
        self.syspath = sysinfo.SYS_FPGA
        self.pcipath = sysinfo.PCI_DEVICES_PATH
        self.pcipath_pattern = sysinfo.PCI_DEVICES_PATH_PATTERN
        tmp_sys_dir = self.useFixture(fixtures.TempDir())
        prepare_test_data.create_fake_sysfs(tmp_sys_dir.path)
        tmp_path = tmp_sys_dir.path
        sysinfo.SYS_FPGA = os.path.join(
            tmp_path, sysinfo.SYS_FPGA.split("/", 1)[-1])
        sysinfo.PCI_DEVICES_PATH = os.path.join(
            tmp_path, sysinfo.PCI_DEVICES_PATH.split("/", 1)[-1])
        sysinfo.PCI_DEVICES_PATH_PATTERN = os.path.join(
            tmp_path, sysinfo.PCI_DEVICES_PATH_PATTERN.split("/", 1)[-1])

    def tearDown(self):
        super(TestIntelFPGADriver, self).tearDown()
        sysinfo.SYS_FPGA = self.syspath
        sysinfo.PCI_DEVICES_PATH = self.pcipath
        sysinfo.PCI_DEVICES_PATH_PATTERN = self.pcipath_pattern

    def test_discover(self):
        attach_handle_list = [
            [
                {'attach_type': 'PCI',
                 'attach_info': '{"bus": "5e", '
                                '"device": "00", '
                                '"domain": "0000", '
                                '"function": "1"}',
                 'in_use': False}
            ],
            [
                {'attach_type': 'PCI',
                 'attach_info': '{"bus": "be", '
                                '"device": "00", '
                                '"domain": "0000", '
                                '"function": "0"}',
                 'in_use': False}
            ]
        ]
        expected = [{'vendor': '0x8086',
                     'type': 'FPGA',
                     'model': '0xbcc0',
                     'deployable_list':
                         [
                             {'num_accelerators': 1,
                              'name': 'intel-fpga-dev_0000:5e:00.1',
                              'attach_handle_list': attach_handle_list[0]
                              },
                         ],
                     'controlpath_id':
                         {
                             'cpid_info': '{"bus": "5e", '
                                          '"device": "00", '
                                          '"domain": "0000", '
                                          '"function": "0"}',
                             'cpid_type': 'PCI'}
                     },
                    {'vendor': '0x8086',
                     'type': 'FPGA',
                     'model': '0xbcc0',
                     'deployable_list':
                         [
                             {'num_accelerators': 1,
                              'name': 'intel-fpga-dev_0000:be:00.0',
                              'attach_handle_list': attach_handle_list[1]
                              },
                         ],
                     'controlpath_id':
                         {
                             'cpid_info': '{"bus": "be", '
                                          '"device": "00", '
                                          '"domain": "0000", '
                                          '"function": "0"}',
                             'cpid_type': 'PCI'}
                     }
                    ]
        intel = IntelFPGADriver()
        fpgas = intel.discover()
        list.sort(fpgas, key=lambda x: x._obj_deployable_list[0].name)
        self.assertEqual(2, len(fpgas))
        for i in range(len(fpgas)):
            fpga_dict = fpgas[i].as_dict()
            fpga_dep_list = fpga_dict['deployable_list']
            fpga_attach_handle_list = \
                fpga_dep_list[0].as_dict()['attach_handle_list']
            self.assertEqual(expected[i]['vendor'], fpga_dict['vendor'])
            self.assertEqual(expected[i]['controlpath_id'],
                             fpga_dict['controlpath_id'])
            self.assertEqual(expected[i]['deployable_list'][0]
                             ['num_accelerators'],
                             fpga_dep_list[0].as_dict()['num_accelerators'])
            self.assertEqual(attach_handle_list[i][0],
                             fpga_attach_handle_list[0].as_dict())

    @mock.patch('cyborg.accelerator.drivers.fpga.intel.driver.'
                '_fpga_program_privileged')
    def test_intel_program(self, mock_prog):
        b = "0x5e"
        d = "0x00"
        f = "0x0"
        expect_cmd = ['--bus', b, '--device', d, '--function', f,
                      '/path/image']

        intel = IntelFPGADriver()
        cpid_info = {"domain": "0000", "bus": "5e",
                     "device": "00", "function": "0"}
        cpid = {'cpid_type': 'PCI', 'cpid_info': cpid_info}

        # program PF
        mock_prog.return_value = bytes([0])
        intel.program(cpid, "/path/image")
        mock_prog.assert_called_with(expect_cmd)



# Copyright 2021 Inspur, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
Cyborg Xilinx FPGA driver implementation.
"""
from oslo_concurrency import processutils

from cyborg.accelerator.drivers.fpga.base import FPGADriver
from cyborg.accelerator.drivers.fpga.xilinx import sysinfo
from cyborg.common import exception
import cyborg.privsep


@cyborg.privsep.sys_admin_pctxt.entrypoint
def _fpga_program_privileged(cmd_args):
    # Example: xbmgmt program --device 0000:d8:00.0
    # --base --image xilinx-u250-gen3x16-base
    cmd = ["/opt/xilinx/xrt/bin/xbmgmt"]
    cmd.extend(cmd_args)
    return processutils.execute(*cmd)


class XilinxFPGADriver(FPGADriver):
    """Class for Xilinx FPGA drivers.
       Vendor should implement their specific drivers in this class.
    """
    VENDOR = "xilinx"

    def __init__(self, *args, **kwargs):
        pass

    def discover(self):
        return sysinfo.fpga_tree()

    def program(self, controlpath_id, image_file_path):
        """Program the FPGA with the provided bitstream image.

        :param controlpath_id: the ID of controlpath.
        :param image_file_path: the path of the image file
        :returns: True on success, False on failure
        """
        if controlpath_id['cpid_type'] != "PCI":
            raise exception.InvalidType(obj='controlpath_id',
                                        type=controlpath_id['cpid_type'],
                                        expected='PCI')
        cmd_args = ['program']
        cmd_args.append('--device')
        bdf_dict = controlpath_id['cpid_info']
        # BDF format: domain:bus:device:function
        bdf = ':'.join([s for s in map(lambda x: bdf_dict[x],
                        ['domain', 'bus', 'device', 'function'])])
        cmd_args.append(bdf)
        cmd_args.append('--base')
        cmd_args.append('--image')
        cmd_args.append(image_file_path)

        try:
            _fpga_program_privileged(cmd_args)
            return True
        except Exception:
            return False



# Copyright 2018 Huawei Technologies Co.,LTD.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from cyborg import objects
from cyborg.objects import fields
from cyborg.objects import physical_function


def fake_db_physical_function(**updates):
    root_uuid = uuidutils.generate_uuid()
    db_physical_function = {
        'id': 1,
        'deleted': False,
        'uuid': root_uuid,
        'name': 'dp_name',
        'parent_uuid': None,
        'root_uuid': root_uuid,
        'address': '00:7f:0b.2',
        'host': 'host_name',
        'board': 'KU115',
        'vendor': 'Xilinx',
        'version': '1.0',
        'type': 'pf',
        'interface_type': 'pci',
        'assignable': True,
        'instance_uuid': None,
        'availability': 'Available',
        'accelerator_id': 1
        }

    for name, field in physical_function.PhysicalFunction.fields.items():
        if name in db_physical_function:
            continue
        if field.nullable:
            db_physical_function[name] = None
        elif field.default != fields.UnspecifiedDefault:
            db_physical_function[name] = field.default
        else:
            raise Exception('fake_db_physical_function needs help with %s'
                            % name)

    if updates:
        db_physical_function.update(updates)

    return db_physical_function


def fake_physical_function_obj(context, obj_pf_class=None, **updates):
    if obj_pf_class is None:
        obj_pf_class = objects.VirtualFunction
    expected_attrs = updates.pop('expected_attrs', None)
    pf = obj_pf_class._from_db_object(context,
                                      obj_pf_class(),
                                      fake_db_physical_function(**updates),
                                      expected_attrs=expected_attrs)
    pf.obj_reset_changes()
    return pf



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import pecan

from oslo_config import cfg
from oslo_log import log
from paste import deploy

from cyborg.api import config
from cyborg.api import hooks
from cyborg.api import middleware
import cyborg.conf


CONF = cyborg.conf.CONF
LOG = log.getLogger(__name__)


def get_pecan_config():
    # Set up the pecan configuration
    filename = config.__file__.replace('.pyc', '.py')
    return pecan.configuration.conf_from_file(filename)


def setup_app(pecan_config=None, extra_hooks=None):
    if not pecan_config:
        pecan_config = get_pecan_config()

    app_hooks = [hooks.ConfigHook(),
                 hooks.ConductorAPIHook(),
                 hooks.ContextHook(pecan_config.app.acl_public_routes),
                 hooks.PublicUrlHook()]
    if extra_hooks:
        app_hooks.extend(extra_hooks)

    app_conf = dict(pecan_config.app)
    app = pecan.make_app(
        app_conf.pop('root'),
        force_canonical=getattr(pecan_config.app, 'force_canonical', True),
        hooks=app_hooks,
        wrap_app=middleware.ParsableErrorMiddleware,
        **app_conf
    )

    return app


def load_app():
    cfg_file = None
    cfg_path = CONF.api.api_paste_config
    if not os.path.isabs(cfg_path):
        cfg_file = CONF.find_file(cfg_path)
    elif os.path.exists(cfg_path):
        cfg_file = cfg_path

    if not cfg_file:
        raise cfg.ConfigFilesNotFoundError([CONF.api.api_paste_config])
    LOG.info("Full WSGI config used: %s", cfg_file)
    return deploy.loadapp("config:" + cfg_file)


def app_factory(global_config, **local_conf):
    return setup_app()



# Copyright 2019 Beijing Lenovo Software Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from openstack import connection
from oslo_log import log as logging
from oslo_utils import versionutils
from oslo_versionedobjects import base as object_base

from cyborg.agent.rpcapi import AgentAPI
from cyborg.common import constants
from cyborg.common.constants import ARQ_STATES_TRANSFORM_MATRIX
from cyborg.common import exception
from cyborg.common import utils
from cyborg.conf import CONF
from cyborg.db import api as dbapi
from cyborg import objects
from cyborg.objects.attach_handle import AttachHandle
from cyborg.objects import base
from cyborg.objects.device_profile import DeviceProfile
from cyborg.objects.extarq.ext_arq_job import ExtARQJobMixin
from cyborg.objects import fields as object_fields


LOG = logging.getLogger(__name__)


@base.CyborgObjectRegistry.register
class ExtARQ(base.CyborgObject, object_base.VersionedObjectDictCompat,
             utils.FactoryMixin, ExtARQJobMixin):
    """ExtARQ is a wrapper around ARQ with Cyborg-private fields.
       Each ExtARQ object contains exactly one ARQ object as a field.
       But, in the db layer, ExtARQ and ARQ are represented together
       as a row in a single table. Both share a single UUID.
       ExtARQ version is bumped up either if any of its fields change
       or if the ARQ version changes.
    """
    # Version 1.0: Initial version
    # 1.1: v2 API and Nova integration
    # 1.2: Fill the value of deployable_id
    VERSION = '1.2'

    dbapi = dbapi.get_instance()

    fields = {
        'arq': object_fields.ObjectField('ARQ'),
        # Cyborg-private fields
        # Left substate open now, fill them out during design/implementation
        # later.
        'substate': object_fields.StringField(),
        'deployable_uuid': object_fields.UUIDField(nullable=True),

        # The dp group is copied in to the extarq, so that any changes or
        # deletions to the device profile do not affect running VMs.
        'device_profile_group': object_fields.DictOfStringsField(
            nullable=True),
        # For bound ARQs, we keep the attach handle ID and deployable ID here
        # so that it is easy to deallocate on unbind or delete.
        'attach_handle_id': object_fields.IntegerField(nullable=True),
        'deployable_id': object_fields.IntegerField(nullable=True),
    }

    def obj_make_compatible(self, primitive, target_version):
        super(ExtARQ, self).obj_make_compatible(
            primitive, target_version)
        target_version = versionutils.convert_version_to_tuple(target_version)
        # TODO(eric): need to handle v1.1 changes
        if target_version < (1, 2) and 'deployable_id' in primitive:
            del primitive['deployable_id']

    def __init__(self, *args, **kwargs):
        super(ExtARQ, self).__init__(*args, **kwargs)
        self.agent = AgentAPI()

    def create(self, context, device_profile_id=None):
        """Create an ExtARQ record in the DB."""
        if 'device_profile_name' not in self.arq and not device_profile_id:
            raise exception.ObjectActionError(
                action='create',
                reason='Device profile name is required in ARQ')
        self.arq.state = constants.ARQ_INITIAL
        self.substate = constants.ARQ_INITIAL
        values = self.obj_get_changes()
        arq_obj = values.pop('arq', None)
        if arq_obj is not None:
            values.update(arq_obj.as_dict())

        # Pass devprof id to db layer, to avoid repeated queries
        if device_profile_id is not None:
            values['device_profile_id'] = device_profile_id

        db_extarq = self.dbapi.extarq_create(context, values)
        self._from_db_object(self, db_extarq, context)
        return self

    @classmethod
    def get(cls, context, uuid, lock=False):
        """Find a DB ExtARQ and return an Obj ExtARQ."""
        db_extarq = cls.dbapi.extarq_get(context, uuid)
        obj_arq = objects.ARQ(context)
        obj_extarq = cls(context)
        obj_extarq['arq'] = obj_arq
        obj_extarq = cls._from_db_object(obj_extarq, db_extarq, context)
        return obj_extarq

    @classmethod
    def list(cls, context, uuid_range=None):
        """Return a list of ExtARQ objects."""
        db_extarqs = cls.dbapi.extarq_list(context, uuid_range)
        obj_extarq_list = cls._from_db_object_list(
            db_extarqs, context)
        return obj_extarq_list

    def save(self, context):
        """Update an ExtARQ record in the DB."""
        updates = self.obj_get_changes()
        db_extarq = self.dbapi.extarq_update(context, self.arq.uuid, updates)
        self._from_db_object(self, db_extarq, context)

    def update_state(self, context, state, scope=None):
        """Update an ExtARQ state record in the DB."""
        updates = self.obj_get_changes()
        updates["state"] = state
        db_extarq = self.dbapi.extarq_update(
            context, self.arq.uuid, updates, scope)
        self._from_db_object(self, db_extarq, context)

    def update_check_state(self, context, state, scope=None):
        if self.arq.state == state:
            LOG.info("ExtARQ(%s) state is %s, no need to update",
                     self.arq.uuid, state)
            return False
        old = self.arq.state
        scope = scope or ARQ_STATES_TRANSFORM_MATRIX[state]
        self.update_state(context, state, scope)
        ea = ExtARQ.get(context, self.arq.uuid, lock=True)
        if not ea:
            raise exception.ResourceNotFound(
                resources='ExtARQ',
                msg="Can not find ExtARQ(%s)" % self.arq.uuid)
        current = ea.arq.state
        if state != current:
            msg = ("Failed to change ARQ state from %s to %s, the current "
                   "state is %s" % (old, state, current))
            LOG.error(msg)
            raise exception.ARQBadState(
                state=current, uuid=self.arq.uuid, expected=list(state))
        return True

    def destroy(self, context):
        """Delete an ExtARQ from the DB."""
        self.dbapi.extarq_delete(context, self.arq.uuid)
        self.obj_reset_changes()

    @classmethod
    def delete_by_uuid(cls, context, arq_uuid_list):
        """Delete a list of ARQs based on their UUIDs.

        This is not idempotent, i.e., if the first call to delete an
        ARQ has succeeded, second and later calls to delete the same ARQ
        will get errored out, but it will raise the exception only after
        all input arq being operated.
        """
        unexisted = []
        for uuid in arq_uuid_list:
            try:
                obj_extarq = objects.ExtARQ.get(context, uuid)
                # TODO() Defer deletion to conductor
                if obj_extarq.arq.state != constants.ARQ_INITIAL:
                    obj_extarq.unbind(context)
                obj_extarq.destroy(context)
            except exception.ResourceNotFound:
                unexisted.append(uuid)
                continue
        if unexisted:
            LOG.warning('There are unexisted arqs: %s', unexisted)
            raise exception.ResourceNotFound(
                resource='ARQ',
                msg='with uuids %s' % unexisted)

    @classmethod
    def delete_by_instance(cls, context, instance_uuid):
        """Delete all ARQs for given instance.

        This is idempotent, i.e., it would have the same effect if called
        repeatedly with the same instance UUID. In other words, it would
        not raise an error on the second and later attempts even if the
        first one has deleted the ARQs.
        """
        obj_extarqs = [extarq for extarq in objects.ExtARQ.list(context)
                       if extarq.arq['instance_uuid'] == instance_uuid]
        for obj_extarq in obj_extarqs:
            LOG.info('Deleting obj_extarq uuid %s for instance %s',
                     obj_extarq.arq['uuid'], obj_extarq.arq['instance_uuid'])
            obj_extarq.unbind(context)
            obj_extarq.destroy(context)

    def _get_glance_connection(self):
        default_user = 'devstack-admin'
        try:
            auth_user = CONF.image.username or default_user
        except Exception:
            auth_user = default_user
        return connection.Connection(cloud=auth_user)

    def _allocate_attach_handle(self, context, deployable):
        try:
            ah = AttachHandle.allocate(context, deployable.id)
            self.attach_handle_id = ah.id
            # if attach_handle is a vgpu, create the mdev in the sys path
            if ah.attach_type == 'MDEV':
                attach_info = json.loads(ah.attach_info)
                pci_addr = "{}:{}:{}.{}".format(
                    attach_info['domain'], attach_info['bus'],
                    attach_info['device'], attach_info['function'])
                hostname = self.arq.hostname
                asked_type = attach_info['asked_type']
                self.agent.create_vgpu_mdev(
                    context, hostname, pci_addr, asked_type, ah.uuid)
        except Exception as e:
            LOG.error("Failed to allocate attach handle for ARQ %s"
                      "from deployable %s. Reason: %s",
                      self.arq.uuid, deployable.uuid, str(e))
            # TODO(Shaohe) Rollback? We have _update_placement,
            # should cancel it.
            self.update_check_state(
                context, constants.ARQ_BIND_FAILED)
            raise
        LOG.info('Attach handle(%s) allocate for ARQ(%s) successfully.',
                 ah.uuid, self.arq.uuid)

    def bind(self, context, deployable):
        self._allocate_attach_handle(context, deployable)
        self.deployable_id = deployable.id
        self.save(context)
        # ARQ state changes get committed here
        self.update_check_state(context, constants.ARQ_BOUND)
        LOG.info('Update ARQ %s state to "Bound" successfully.',
                 self.arq.uuid)
        # TODO(Shaohe) rollback self._unbind and self._delete
        # if (self.arq.state == constants.ARQ_DELETING
        #         or self.arq.state == ARQ_UNBOUND):

    def _deallocate_attach_handle(self, context, ah_id, hostname):
        try:
            attach_handle = AttachHandle.get_by_id(context, ah_id)
            if attach_handle.attach_type == 'MDEV':
                attach_info = json.loads(attach_handle.attach_info)
                pci_addr = "{}:{}:{}.{}".format(
                    attach_info['domain'], attach_info['bus'],
                    attach_info['device'], attach_info['function'])
                self.agent.remove_vgpu_mdev(
                    context, hostname, pci_addr,
                    attach_info['asked_type'], attach_handle.uuid)
            attach_handle.deallocate(context)
        except Exception as e:
            LOG.error("Failed to deallocate attach handle %s for ARQ %s."
                      "Reason: %s", ah_id, self.arq.uuid, str(e))
            self.update_check_state(
                context, constants.ARQ_UNBIND_FAILED)
            raise
        LOG.info('Attach handle(%s) deallocate for ARQ(%s) successfully.',
                 ah_id, self.arq.uuid)

    def unbind(self, context):
        arq = self.arq
        hostname = arq.hostname
        arq.hostname = None
        arq.device_rp_uuid = None
        arq.instance_uuid = None
        arq.state = constants.ARQ_UNBOUND

        # Unbind: mark attach handles as freed
        ah_id = self.attach_handle_id
        if ah_id:
            self._deallocate_attach_handle(context, ah_id, hostname)
        self.attach_handle_id = None
        self.deployable_id = None
        self.save(context)

    @classmethod
    def _fill_obj_extarq_fields(cls, context, db_extarq):
        """ExtARQ object has some fields that are not present
           in db_extarq. We fill them out here.
        """
        # From the 2 fields in the ExtARQ, we obtain other fields.
        devprof_id = db_extarq['device_profile_id']
        devprof_group_id = db_extarq['device_profile_group_id']

        devprof = DeviceProfile.get_by_id(context, devprof_id)
        db_extarq['device_profile_name'] = devprof['name']

        db_extarq['attach_handle_type'] = ''
        db_extarq['attach_handle_info'] = ''
        if db_extarq['state'] == 'Bound':  # TODO() Do proper bind
            db_ah = cls.dbapi.attach_handle_get_by_id(
                context, db_extarq['attach_handle_id'])
            if db_ah is not None:
                db_extarq['attach_handle_type'] = db_ah['attach_type']
                db_extarq['attach_handle_info'] = db_ah['attach_info']
                db_extarq['attach_handle_uuid'] = db_ah['uuid']
            else:
                raise exception.ResourceNotFound(
                    resource='Attach Handle',
                    msg='with uuid=%s' % db_extarq['attach_handle_id'])

        if db_extarq['deployable_id']:
            dep = objects.Deployable.get_by_id(context,
                                               db_extarq['deployable_id'])
            db_extarq['deployable_uuid'] = dep.uuid
        else:
            LOG.debug('Setting deployable UUID to zeroes for db_extarq %s',
                      db_extarq['uuid'])
            db_extarq['deployable_uuid'] = (
                '00000000-0000-0000-0000-000000000000')

        groups = devprof['groups']
        db_extarq['device_profile_group'] = groups[devprof_group_id]

        return db_extarq

    @classmethod
    def _from_db_object(cls, extarq, db_extarq, context):
        """Converts an ExtARQ to a formal object.
        :param extarq: An object of the class ExtARQ
        :param db_extarq: A DB model of the object
        :return: The object of the class with the database entity added
        """
        cls._fill_obj_extarq_fields(context, db_extarq)

        for field in extarq.fields:
            if field != 'arq':
                extarq[field] = db_extarq.get(field)
        extarq.arq = objects.ARQ()
        extarq.arq._from_db_object(extarq.arq, db_extarq)
        extarq.obj_reset_changes()
        return extarq

    @classmethod
    def _from_db_object_list(cls, db_objs, context):
        """Converts a list of ExtARQs to a list of formal objects."""
        objs = []
        for db_obj in db_objs:
            extarq = cls(context)
            obj = cls._from_db_object(extarq, db_obj, context)
            objs.append(obj)
        return objs

    def obj_get_changes(self):
        """Returns a dict of changed fields and their new values."""
        changes = {}
        for key in self.obj_what_changed():
            if key != 'arq':
                changes[key] = getattr(self, key)

        for key in self.arq.obj_what_changed():
            changes[key] = getattr(self.arq, key)

        return changes



"""add_nic_type

Revision ID: 899cead40bc9
Revises: 7e6f1f107f2b
Create Date: 2020-09-18 02:33:42.640673

"""

# revision identifiers, used by Alembic.
revision = '899cead40bc9'
down_revision = '7e6f1f107f2b'

from alembic import op
import sqlalchemy as sa


def upgrade():
    new_device_type = sa.Enum('GPU', 'FPGA', 'AICHIP', 'QAT', 'NIC',
                             name='device_type')
    op.alter_column('devices', 'type',
                    existing_type=new_device_type,
                    nullable=False)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Main abstraction layer for retrieving and storing information about accelerator
images used by the cyborg agent layer.
"""

from cyborg.image import glance
from oslo_log import log

LOG = log.getLogger(__name__)


class API(object):

    """Responsible for exposing a relatively stable internal API for other
    modules in Cyborg to retrieve information about accelerator images.
    """

    def _get_session_and_image_id(self, context, id_or_uri):
        """Returns a tuple of (session, image_id). If the supplied `id_or_uri`
        is an image ID, then the default client session will be returned
        for the context's user, along with the image ID. If the supplied
        `id_or_uri` parameter is a URI, then a client session connecting to
        the URI's image service endpoint will be returned along with a
        parsed image ID from that URI.

        :param context: The `cyborg.context.Context` object for the request
        :param id_or_uri: A UUID identifier or an image URI to look up image
                          information for.
        """
        return glance.get_remote_image_service(context, id_or_uri)

    def download(self, context, id_or_uri, data=None, dest_path=None):
        """Transfer image bits from Glance or a known source location to the
        supplied destination filepath.

        :param context: The `cyborg.context.RequestContext` object for the
                        request
        :param id_or_uri: A UUID identifier or an image URI to look up image
                          information for.
        :param data: A file object to use in downloading image data.
        :param dest_path: Filepath to transfer image bits to.

        Note that because of the poor design of the
        `glance.ImageService.download` method, the function returns different
        things depending on what arguments are passed to it. If a data argument
        is supplied but no dest_path is specified (only done in the XenAPI virt
        driver's image.utils module) then None is returned from the method. If
        the data argument is not specified but a destination path *is*
        specified, then a writeable file handle to the destination path is
        constructed in the method and the image bits written to that file, and
        again, None is returned from the method. If no data argument is
        supplied and no dest_path argument is supplied (VMWare and XenAPI virt
        drivers), then the method returns an iterator to the image bits that
        the caller uses to write to wherever location it wants.

        I think the above points to just how hacky/wacky all of this code is,
        and the reason it needs to be cleaned up and standardized across the
        virt driver callers.
        """

        session, image_id = self._get_session_and_image_id(context, id_or_uri)
        return session.download(context, image_id, data=data,
                                dst_path=dest_path)



# Copyright 2020 Intel, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from cyborg.tests.unit.api.controllers.v2 import base as v2_test
from cyborg.tests.unit import fake_deployable


class TestDeployablesController(v2_test.APITestV2):

    DEPLOYABLE_URL = '/deployables'

    def setUp(self):
        super(TestDeployablesController, self).setUp()
        self.headers = self.gen_headers(self.context)
        self.fake_deployable = fake_deployable.fake_deployable_obj(
            self.context)

    def _validate_links(self, links, deployable_uuid):
        has_self_link = False
        for link in links:
            if link['rel'] == 'self':
                has_self_link = True
                url = link['href']
                components = url.split('/')
                self.assertEqual(components[-1], deployable_uuid)
        self.assertTrue(has_self_link)

    def _validate_deployable(self, in_deployable, out_deployable):
        for field in in_deployable.keys():
            if field != 'id':
                self.assertEqual(in_deployable[field], out_deployable[field])
        # Check that the link is properly set up
        self._validate_links(out_deployable['links'], in_deployable['uuid'])

    @mock.patch('cyborg.objects.Deployable.get')
    def test_get_one_by_uuid(self, mock_deployable):
        in_deployable = self.fake_deployable
        mock_deployable.return_value = in_deployable
        uuid = in_deployable['uuid']

        url = self.DEPLOYABLE_URL + '/%s'
        out_deployable = self.get_json(url % uuid, headers=self.headers)
        mock_deployable.assert_called_once()
        self._validate_deployable(in_deployable, out_deployable)

    @mock.patch('cyborg.objects.Deployable.list')
    def test_get_all(self, mock_deployables):
        mock_deployables.return_value = [self.fake_deployable]
        data = self.get_json(self.DEPLOYABLE_URL, headers=self.headers)
        out_deployable = data['deployables']
        self.assertIsInstance(out_deployable, list)
        for out_dev in out_deployable:
            self.assertIsInstance(out_dev, dict)
        self.assertEqual(len(out_deployable), 1)
        self._validate_deployable(self.fake_deployable, out_deployable[0])

    @mock.patch('cyborg.objects.Deployable.list')
    def test_get_with_filters(self, mock_deployables):
        mock_deployables.return_value = [self.fake_deployable]
        # TODO(Xinran) Add API doc to explain the usage of filter.
        # Add "?filters.field=limit&filters.value=1" in DEPLOYABLE_URL, in
        # order to list the deployables with limited number which is 1.
        data = self.get_json(
            self.DEPLOYABLE_URL + "?filters.field=limit&filters.value=1",
            headers=self.headers)
        out_deployable = data['deployables']
        mock_deployables.assert_called_once_with(mock.ANY,
                                                 filters={"limit": "1"})
        self._validate_deployable(self.fake_deployable, out_deployable[0])

    @mock.patch('cyborg.objects.Deployable.list')
    def test_get_with_filters_not_match(self, mock_deployables):
        # This will return null list because the fake deployable's name
        # is "dp_name".
        mock_deployables.return_value = []
        data = self.get_json(
            self.DEPLOYABLE_URL +
            "?filters.field=name&filters.value=wrongname",
            headers=self.headers)
        out_deployable = data['deployables']
        mock_deployables.assert_called_once_with(mock.ANY,
                                                 filters={"name": "wrongname"})
        self.assertEqual(len(out_deployable), 0)



# Copyright 2020 ZTE Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_log import versionutils
from oslo_policy import policy

from cyborg.policies import base


# NOTE(yumeng)During the Policy-default-refresh work, the old device_profile
# policies will be marked as deprecated device_profile policies.
# To ensure API works fine with both old policies and new policies, we set
# ``cyborg.conf [oslo_policy] enforce_scope = False`` by default. With this,
# policy authorization check will pass those who comply with either new policy
# rules or old policy rules by invoking oslo_policy.policy.OrCheck
# (REF:https://github.com/openstack/oslo.policy/blob/cab28649c689067970a51a2f9b329bdd6a0f0501/oslo_policy/policy.py#L726)
# And once we move to new defaults only world, we will set
# ``cyborg.conf [oslo_policy] enforce_scope = True`` by default, at which time
# we can totally remove these deprecated device_profile policies from code.
deprecated_get_all = policy.DeprecatedRule(
    name='cyborg:device_profile:get_all',
    check_str=base.deprecated_default,
    deprecated_reason=('request admin_or_owmer rule is too strict for '
                       'listing device_profile'),
    deprecated_since=versionutils.deprecated.WALLABY)
deprecated_get_one = policy.DeprecatedRule(
    name='cyborg:device_profile:get_one',
    check_str=base.deprecated_default,
    deprecated_reason=('request admin_or_owmer rule is too strict for '
                       'retrieving a device_profile'),
    deprecated_since=versionutils.deprecated.WALLABY)
deprecated_create = policy.DeprecatedRule(
    name='cyborg:device_profile:create',
    check_str=base.deprecated_is_admin,
    deprecated_reason=('project_admin_or_owner is too permissive, '
                       'introduce admin for creation'),
    deprecated_since=versionutils.deprecated.WALLABY)
deprecated_delete = policy.DeprecatedRule(
    name='cyborg:device_profile:delete',
    check_str=base.deprecated_default,
    deprecated_reason=('project_admin_or_owner is too permissive, '
                       'introduce admin for deletion'),
    deprecated_since=versionutils.deprecated.WALLABY)

# new device_profile policies
device_profile_policies = [
    policy.DocumentedRuleDefault(
        name='cyborg:device_profile:get_all',
        check_str=base.PROJECT_READER_OR_ADMIN,
        description='Retrieve all device_profiles',
        operations=[
            {
                'path': '/v2/device_profiles',
                'method': 'GET'
            }],
        scope_types=['project'],
        deprecated_rule=deprecated_get_all),
    policy.DocumentedRuleDefault(
        name='cyborg:device_profile:get_one',
        check_str=base.PROJECT_READER_OR_ADMIN,
        description='Retrieve a specific device_profile',
        operations=[
            {
                'path': '/v2/device_profiles/{device_profiles_uuid}',
                'method': 'GET'
            }],
        scope_types=['project'],
        deprecated_rule=deprecated_get_one),
    policy.DocumentedRuleDefault(
        name='cyborg:device_profile:create',
        check_str=base.ADMIN,
        description='Create a device_profile',
        operations=[
            {
                'path': '/v2/device_profiles',
                'method': 'POST'
            }],
        scope_types=['project'],
        deprecated_rule=deprecated_create),
    policy.DocumentedRuleDefault(
        name='cyborg:device_profile:delete',
        check_str=base.ADMIN,
        description='Delete device_profile(s)',
        operations=[
            {
                'path': '/v2/device_profiles/{device_profiles_uuid}',
                'method': 'DELETE'},
            {
                'path': '/v2/device_profiles?value={device_profile_name1}',
                'method': 'DELETE'},
            ],
        scope_types=['project'],
        deprecated_rule=deprecated_delete),
]


def list_policies():
    return device_profile_policies



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from alembic import script
import contextlib
import fixtures
import sqlalchemy
from unittest import mock

from oslo_db import exception as db_exc
from oslo_db.sqlalchemy import enginefacade
from oslo_db.sqlalchemy import test_fixtures
from oslo_db.sqlalchemy import utils as db_utils
from oslo_log import log as logging
from oslotest import base as test_base

from cyborg.db.sqlalchemy import migration
from cyborg.tests import base

LOG = logging.getLogger(__name__)

# NOTE(brinzhang): This was introduced after migration tests started taking
# more time in gate. Timeout value in seconds for tests performing migrations.
MIGRATIONS_TIMEOUT = 300


@contextlib.contextmanager
def patch_with_engine(engine):
    with mock.patch.object(enginefacade.writer,
                           'get_engine') as patch_engine:
        patch_engine.return_value = engine
        yield


class WalkVersionsMixin(object):
    def _walk_versions(self, engine=None, alembic_cfg=None):
        # Determine latest version script from the repo, then
        # upgrade from 1 through to the latest, with no data
        # in the databases. This just checks that the schema itself
        # upgrades successfully.

        # Place the database under version control
        with patch_with_engine(engine):
            script_directory = script.ScriptDirectory.from_config(alembic_cfg)

            self.assertIsNone(self.migration_api.version(alembic_cfg))

            versions = [ver for ver in script_directory.walk_revisions()]

            for version in reversed(versions):
                self._migrate_up(engine, alembic_cfg,
                                 version.revision, with_data=True)

    def _skippable_migrations(self):
        # Some db scripts are not necessary to check
        special = []
        return special

    def _migrate_up(self, engine, config, version, with_data=False):
        """migrate up to a new version of the db.

        We allow for data insertion and post checks at every
        migration version with special _pre_upgrade_### and
        _check_### functions in the main test
        """
        self.migration_api.upgrade(version, config=config)
        self.assertEqual(version, self.migration_api.version(config))
        if with_data:
            check = getattr(self, "_check_%s" % version, None)
            if version not in self._skippable_migrations():
                self.assertIsNotNone(check, ('DB Migration %i does not have '
                                     'a test. Please add one!') % version)


class TestWalkVersions(base.TestCase, WalkVersionsMixin):
    def setUp(self):
        super(TestWalkVersions, self).setUp()
        self.migration_api = mock.MagicMock()
        self.engine = mock.MagicMock()
        self.config = mock.MagicMock()
        self.versions = [mock.Mock(revision='2b2'), mock.Mock(revision='1a1')]

    def test_migrate_up(self):
        self.migration_api.version.return_value = '6a7f90fc3s8c'
        self._migrate_up(self.engine, self.config, '6a7f90fc3s8c')
        self.migration_api.upgrade.assert_called_with('6a7f90fc3s8c',
                                                      config=self.config)
        self.migration_api.version.assert_called_with(self.config)


class CyborgMigrationsCheckers(object):

    def setUp(self):
        super(CyborgMigrationsCheckers, self).setUp()
        self.engine = enginefacade.writer.get_engine()
        self.config = migration._alembic_config()
        self.migration_api = migration
        self.useFixture(fixtures.Timeout(MIGRATIONS_TIMEOUT,
                                         gentle=True)),

    def test_walk_versions(self):
        self._walk_versions(self.engine, self.config)

    def _check_589ff20545b7(self, engine, data):
        devices = db_utils.get_table(engine, 'devices')
        col_names = [column.name for column in devices.c]
        self.assertIn('type', col_names)
        self.assertIsInstance(devices.c.type.type,
                              sqlalchemy.types.Enum)

    def test_upgrade_and_version(self):
        with patch_with_engine(self.engine):
            self.migration_api.upgrade('head')
            self.assertIsNotNone(self.migration_api.version())

    def test_create_schema_and_version(self):
        with patch_with_engine(self.engine):
            self.migration_api.create_schema()
            self.assertIsNotNone(self.migration_api.version())

    def test_upgrade_and_create_schema(self):
        with patch_with_engine(self.engine):
            self.migration_api.upgrade('ede4e3f1a232')
            self.assertRaises(db_exc.DBMigrationError,
                              self.migration_api.create_schema)

    def test_upgrade_twice(self):
        with patch_with_engine(self.engine):
            self.migration_api.upgrade('ede4e3f1a232')
            v1 = self.migration_api.version()
            self.migration_api.upgrade('head')
            v2 = self.migration_api.version()
            self.assertNotEqual(v1, v2)


class TestCyborgMigrationsMySQL(CyborgMigrationsCheckers,
                                WalkVersionsMixin,
                                test_fixtures.OpportunisticDBTestMixin,
                                test_base.BaseTestCase):
    FIXTURE = test_fixtures.MySQLOpportunisticFixture


class TestMigrationsPostgreSQL(CyborgMigrationsCheckers,
                               WalkVersionsMixin,
                               test_fixtures.OpportunisticDBTestMixin,
                               test_base.BaseTestCase):
    FIXTURE = test_fixtures.PostgresqlOpportunisticFixture



# Copyright 2020 Inspur Electronic Information Industry Co.,LTD.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
Cyborg Generic SSD driver implementation.
"""
from oslo_log import log as logging

from cyborg.accelerator.common import utils as pci_utils
from cyborg.accelerator.drivers.ssd import utils

LOG = logging.getLogger(__name__)
VENDOR_MAPS = pci_utils.get_vendor_maps()


class SSDDriver(object):
    """Generic class for SSD drivers.

       This is just a virtual SSD drivers interface.
       Vendor should implement their specific drivers.
    """

    @classmethod
    def create(cls, vendor=None, *args, **kwargs):
        if not vendor:
            return cls(*args, **kwargs)
        for sclass in cls.__subclasses__():
            vendor_name = VENDOR_MAPS.get(vendor, vendor)
            if vendor_name == sclass.VENDOR:
                return sclass(*args, **kwargs)
        raise LookupError("Not find the SSD driver for vendor %s" % vendor)

    def discover(self):
        """Discover SSD information of current vendor(Identified by class).

           If no vendor-specific driver provided, this will return all NVMe SSD
           devices
        :return: List of SSD information dict.
        """
        LOG.info('The method "discover" is called in generic.SSDDriver')
        devs = utils.discover_ssds()
        return devs

    @classmethod
    def discover_vendors(cls):
        """Discover SSD vendors of current node.

        :return: SSD vendor ID list.
        """
        return utils.discover_vendors()



# Copyright 2020 Inspur Electronic Information Industry Co.,LTD..
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from cyborg.accelerator.drivers.ssd.base import SSDDriver
from cyborg.accelerator.drivers.ssd.inspur.driver import InspurNVMeSSDDriver
from cyborg.tests import base


class TestSSDDriver(base.TestCase):
    def test_create_base_ssd_driver(self):
        drv = SSDDriver.create()
        self.assertIsInstance(drv, SSDDriver)
        self.assertNotIsInstance(drv, InspurNVMeSSDDriver)

    def test_create_inspur_ssd_driver(self):
        drv = SSDDriver.create(vendor='inspur')
        self.assertIsInstance(drv, InspurNVMeSSDDriver)

    def test_create_ssd_vendor_not_found(self):
        self.assertRaises(LookupError, SSDDriver.create, '_non-exist_vendor')



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""update_for_nova_integ

Revision ID: c1b5abada09c
Revises: 589ff20545b7
Create Date: 2019-07-11 22:13:45.773499

"""

# revision identifiers, used by Alembic.
revision = 'c1b5abada09c'
down_revision = '589ff20545b7'

from alembic import op
import sqlalchemy as sa

from cyborg.common import constants


def upgrade():
    # Update Deployables
    op.add_column(
        'deployables',
        sa.Column('rp_uuid', sa.String(length=36), nullable=True))
    op.add_column(
        'deployables',
        sa.Column('driver_name', sa.String(length=100), nullable=True))
    op.add_column(
        'deployables',
        sa.Column('bitstream_id', sa.String(length=36), nullable=True))

    # Update ExtARQ table
    op.add_column(
        'extended_accelerator_requests',
        sa.Column('device_profile_group_id', sa.Integer(), nullable=False))
    op.add_column(
        'extended_accelerator_requests',
        sa.Column('instance_uuid', sa.String(length=36),
                  nullable=True))
    op.create_index('extArqs_instance_uuid_idx',  # index name
                    'extended_accelerator_requests',  # table name
                    ['instance_uuid']  # columns on which index is defined
                    )
    op.drop_index('extArqs_device_instance_uuid_idx',  # index name
                  'extended_accelerator_requests',  # table name
                  )
    op.drop_column('extended_accelerator_requests', 'device_instance_uuid')
    # Add more valid states for 'state' field
    ns = sa.Enum(constants.ARQ_INITIAL,
                 constants.ARQ_BIND_STARTED,
                 constants.ARQ_BOUND,
                 constants.ARQ_UNBOUND,
                 constants.ARQ_BIND_FAILED,
                 constants.ARQ_DELETING, name='state')
    op.alter_column(
        'extended_accelerator_requests', 'state',
        existing_type=ns, nullable=False, default=constants.ARQ_INITIAL)

    # update attach type fields
    new_attach_type = sa.Enum(constants.AH_TYPE_PCI,
                              constants.AH_TYPE_MDEV,
                              constants.AH_TYPE_TEST_PCI,
                              name='attach_type')
    op.alter_column('attach_handles', 'attach_type',
                    existing_type=new_attach_type,
                    nullable=False)

    # Update device_profiles table to make name and uuid unique separately.
    # Previous schema made the pair unique.
    op.create_unique_constraint('uniq_device_profiles0uuid',
                                'device_profiles', ['uuid'])
    op.create_unique_constraint('uniq_device_profiles0name',
                                'device_profiles', ['name'])
    op.drop_constraint('uniq_device_profiles0uuid0name',
                       'device_profiles', type_='unique')



# Copyright 2023 Inspur, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from cyborg.accelerator.common import utils
import unittest


class TestUtils(unittest.TestCase):
    def setUp(self):
        self.utils = utils
        super(TestUtils, self).setUp()

    def test_pci_str_to_json(self):
        pci_address = '0000:0b:00.0'
        json_str = '{"bus": "0b", "device": "00", "domain": "0000", ' \
                   '"function": "0"}'
        result = self.utils.pci_str_to_json(pci_address)
        self.assertEqual(result, json_str)

        pci_address = '0000:0b:00.1'
        json_str = '{"bus": "0b", "device": "00", "domain": "0000", ' \
                   '"function": "1", "physical_network": "physnet"}'
        result = self.utils.pci_str_to_json(pci_address, 'physnet')
        self.assertEqual(result, json_str)

    def test_mdev_str_to_json(self):
        json_str = '{"asked_type": "type", "bus": "0b", "device": "00", ' \
                   '"domain": "0000", "function": "1", "vgpu_mark": "mask"}'
        result = self.utils.mdev_str_to_json('0000:0b:00.1', 'type', 'mask')
        self.assertEqual(result, json_str)

        json_str = '{"asked_type": null, "bus": "0b", "device": "00", ' \
                   '"domain": "0000", "function": "1", "vgpu_mark": null}'
        result = self.utils.mdev_str_to_json('0000:0b:00.1', None, None)
        self.assertEqual(result, json_str)

    def test_parse_address(self):
        result = self.utils.parse_address('0000:0b:00.1')
        self.assertEqual(result, ('0000', '0b', '00', '1'))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Utils for SPDK driver.
"""

import glob
import os
import re

from oslo_config import cfg
from oslo_log import log as logging

from cyborg.accelerator.common import exception
from cyborg.accelerator import configuration
from cyborg.accelerator.drivers.spdk.util.pyspdk.nvmf_client import NvmfTgt
from cyborg.accelerator.drivers.spdk.util.pyspdk.py_spdk import PySPDK
from cyborg.accelerator.drivers.spdk.util.pyspdk.vhost_client import VhostTgt
from cyborg.common.i18n import _

LOG = logging.getLogger(__name__)

accelerator_opts = [
    cfg.StrOpt('spdk_conf_file',
               default='/etc/cyborg/spdk.conf',
               help=_('SPDK conf file to be used for the SPDK driver')),

    cfg.StrOpt('accelerator_servers',
               default=['vhost', 'nvmf', 'iscsi'],
               help=_('A list of accelerator servers to enable by default')),

    cfg.StrOpt('spdk_dir',
               default='/home/wewe/spdk',
               help=_('The SPDK directory is /home/{user_name}/spdk')),

    cfg.StrOpt('device_type',
               default='NVMe',
               help=_('Backend device type is NVMe by default')),

    cfg.BoolOpt('remoteable',
                default=False,
                help=_('Remoteable is false by default'))
]

CONF = cfg.CONF
CONF.register_opts(accelerator_opts, group=configuration.SHARED_CONF_GROUP)

config = configuration.Configuration(accelerator_opts)
config.append_config_values(accelerator_opts)
SERVERS = config.safe_get('accelerator_servers')
SERVERS_PATTERN = re.compile("|".join(["(%s)" % s for s in SERVERS]))
SPDK_SERVER_APP_DIR = os.path.join(config.safe_get('spdk_dir'), 'app/')


def discover_servers():
    """Discover backend servers according to the CONF

    :returns: server list.
    """
    servers = set()
    for p in glob.glob1(SPDK_SERVER_APP_DIR, "*"):
        m = SERVERS_PATTERN.match(p)
        if m:
            servers.add(m.group())
    return list(servers)


def delete_bdev(py, accelerator, name):
    """Delete a blockdev

    :param py: py_client.
    :param accelerator: accelerator.
    :param name: Blockdev name to be deleted.
    """
    acc_client = get_accelerator_client(py, accelerator)
    acc_client.delete_bdev(name)


def kill_instance(py, accelerator, sig_name):
    """Send signal to instance

    :param py: py_client.
    :param accelerator: accelerator.
    :param sig_name: signal will be sent to server.
    """
    acc_client = get_accelerator_client(py, accelerator)
    acc_client.kill_instance(sig_name)


def construct_aio_bdev(py, accelerator, filename, name, block_size):
    """Add a bdev with aio backend

    :param py: py_client.
    :param accelerator: accelerator.
    :param filename: Path to device or file (ex: /dev/sda).
    :param name: Block device name.
    :param block_size: Block size for this bdev.
    :return: name.
    """
    acc_client = get_accelerator_client(py, accelerator)
    acc_client.construct_aio_bdev(filename, name, block_size)
    return name


def construct_error_bdev(py, accelerator, basename):
    """Add a bdev with error backend

    :param py: py_client.
    :param accelerator: accelerator.
    :param basename: Path to device or file (ex: /dev/sda).
    """
    acc_client = get_accelerator_client(py, accelerator)
    acc_client.construct_error_bdev(basename)


def construct_nvme_bdev(py,
                        accelerator,
                        name,
                        trtype,
                        traddr,
                        adrfam,
                        trsvcid,
                        subnqn
                        ):
    """Add a bdev with nvme backend

    :param py: py_client.
    :param accelerator: accelerator.
    :param name: Name of the bdev.
    :param trtype: NVMe-oF target trtype: e.g., rdma, pcie.
    :param traddr: NVMe-oF target address: e.g., an ip address
    or BDF.
    :param adrfam: NVMe-oF target adrfam: e.g., ipv4, ipv6, ib,
    fc, intra_host.
    :param trsvcid: NVMe-oF target trsvcid: e.g., a port number.
    :param subnqn: NVMe-oF target subnqn.
    :return: name.
    """
    acc_client = get_accelerator_client(py, accelerator)
    acc_client.construct_nvme_bdev(name,
                                   trtype,
                                   traddr,
                                   adrfam,
                                   trsvcid,
                                   subnqn
                                   )
    return name


def construct_null_bdev(py,
                        accelerator,
                        name,
                        total_size,
                        block_size
                        ):
    """Add a bdev with null backend

    :param py: py_client.
    :param accelerator: accelerator.
    :param name: Block device name.
    :param total_size: Size of null bdev in MB (int > 0).
    :param block_size: Block size for this bdev.
    :return: name.
    """
    acc_client = get_accelerator_client(py, accelerator)
    acc_client.construct_null_bdev(name, total_size, block_size)
    return name


def get_py_client(server):
    """Get the py_client instance

    :param server: server.
    :return: Boolean.
    :raise: InvalidAccelerator.
    """
    if server in SERVERS:
        py = PySPDK(server)
        return py
    else:
        msg = (_("Could not find %s accelerator") % server)
        raise exception.InvalidAccelerator(msg)


def check_for_setup_error(py, server):
    """Check server's status

    :param py: py_client.
    :param server: server.
    :return: Boolean.
    :raise: AcceleratorException.
    """
    if py.is_alive():
        return True
    else:
        msg = (_("%s accelerator is down") % server)
        raise exception.AcceleratorException(msg)


def get_accelerator_client(py, accelerator):
    """Get the specific client that communicates with server

    :param py: py_client.
    :param accelerator: accelerator.
    :return: acc_client.
    :raise: InvalidAccelerator.
    """
    acc_client = None
    if accelerator == 'vhost':
        acc_client = VhostTgt(py)
        return acc_client
    elif accelerator == 'nvmf':
        acc_client = NvmfTgt(py)
        return acc_client
    else:
        exc_msg = (_("accelerator_client %(acc_client) is missing")
                   % acc_client)
        raise exception.InvalidAccelerator(exc_msg)



# Copyright 2019 Beijing Lenovo Software Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from testtools.matchers import HasLength

from cyborg.common import constants
from cyborg.common import exception
from cyborg import objects
from cyborg.tests.unit.db import base
from cyborg.tests.unit import fake_attach_handle
from cyborg.tests.unit import fake_deployable
from cyborg.tests.unit import fake_device_profile
from cyborg.tests.unit import fake_extarq


class TestExtARQObject(base.DbTestCase):

    def setUp(self):
        super(TestExtARQObject, self).setUp()
        self.fake_db_extarqs = fake_extarq.get_fake_db_extarqs()
        self.fake_obj_extarqs = fake_extarq.get_fake_extarq_objs()
        self.fake_obj_fpga_extarqs = fake_extarq.get_fake_fpga_extarq_objs()
        self.fake_obj_ahs = fake_attach_handle.get_fake_attach_handle_objs()
        self.deployable_uuids = ['0acbf8d6-e02a-4394-aae3-57557d209498']

    @mock.patch('cyborg.objects.ExtARQ._from_db_object')
    def test_get(self, mock_from_db_obj):
        db_extarq = self.fake_db_extarqs[0]
        uuid = db_extarq['uuid']
        mock_from_db_obj.return_value = self.fake_obj_extarqs[0]
        with mock.patch.object(self.dbapi, 'extarq_get',
                               autospec=True) as mock_extarq_get:
            mock_extarq_get.return_value = db_extarq
            obj_extarq = objects.ExtARQ.get(self.context, uuid)
            mock_extarq_get.assert_called_once_with(self.context, uuid)
            self.assertEqual(obj_extarq.arq.uuid, uuid)

    @mock.patch('cyborg.objects.ExtARQ._from_db_object')
    def test_list(self, mock_from_db_obj):
        db_extarq = self.fake_db_extarqs[0]
        mock_from_db_obj.return_value = self.fake_obj_extarqs[0]
        with mock.patch.object(self.dbapi, 'extarq_list',
                               autospec=True) as mock_get_list:
            mock_get_list.return_value = [db_extarq]
            obj_extarqs = objects.ExtARQ.list(self.context)
            self.assertEqual(1, mock_get_list.call_count)
            self.assertThat(obj_extarqs, HasLength(1))
            self.assertIsInstance(obj_extarqs[0], objects.ExtARQ)
            for obj_extarq in obj_extarqs:
                self.assertEqual(obj_extarqs[0].arq.uuid, db_extarq['uuid'])

    @mock.patch('cyborg.objects.ExtARQ._from_db_object')
    def test_create(self, mock_from_db_obj):
        db_extarq = self.fake_db_extarqs[0]
        mock_from_db_obj.return_value = self.fake_obj_extarqs[0]
        with mock.patch.object(self.dbapi, 'extarq_create',
                               autospec=True) as mock_extarq_create:
            mock_extarq_create.return_value = db_extarq
            extarq = objects.ExtARQ(self.context, **db_extarq)
            extarq.arq = objects.ARQ(self.context, **db_extarq)
            extarq.create(self.context)
            mock_extarq_create.assert_called_once()

    @mock.patch('openstack.connection.Connection')
    @mock.patch('cyborg.common.nova_client.NovaAPI.notify_binding')
    @mock.patch('cyborg.objects.ExtARQ.bind')
    @mock.patch('cyborg.objects.ExtARQ.get')
    def test_apply_patch_to_bad_arq_state(
            self, mock_get, mock_bind, mock_notify_bind, mock_conn):
        good_states = constants.ARQ_STATES_TRANSFORM_MATRIX[
            constants.ARQ_BIND_STARTED]
        mock_get.return_value = obj_extarq = self.fake_obj_extarqs[0]
        uuid = obj_extarq.arq.uuid
        instance_uuid = obj_extarq.arq.instance_uuid
        valid_fields = {
            uuid: {'hostname': obj_extarq.arq.hostname,
                   'device_rp_uuid': obj_extarq.arq.device_rp_uuid,
                   'instance_uuid': instance_uuid}
            }
        patch_list = {
            str(uuid): [
                {"path": "/hostname", "op": "add",
                 "value": obj_extarq.arq.hostname},
                {"path": "/device_rp_uuid", "op": "add",
                 "value": obj_extarq.arq.device_rp_uuid},
                {"path": "/instance_uuid", "op": "add",
                 "value": instance_uuid}
            ]
        }

        for state in set(constants.ARQ_STATES) - set(good_states):
            obj_extarq.arq.state = state
            mock_get.return_value = obj_extarq
            self.assertRaises(
                exception.ARQBadState, objects.ExtARQ.apply_patch,
                self.context, patch_list, valid_fields)

        mock_notify_bind.assert_not_called()

    @mock.patch('cyborg.objects.ExtARQ.save')
    @mock.patch('cyborg.objects.extarq.ext_arq_job.ExtARQJobMixin.'
                'get_arq_bind_statuses')
    @mock.patch('openstack.connection.Connection')
    @mock.patch('cyborg.common.nova_client.NovaAPI.notify_binding')
    @mock.patch('cyborg.objects.ExtARQ._allocate_attach_handle')
    @mock.patch('cyborg.objects.ExtARQ.get')
    @mock.patch('cyborg.objects.ExtARQ.list')
    @mock.patch('cyborg.objects.ExtARQ.update_check_state')
    @mock.patch('cyborg.objects.deployable.Deployable.get_by_device_rp_uuid')
    def test_apply_patch_for_common_extarq(
        self, mock_get_dep, mock_check_state, mock_list, mock_get,
        mock_attach_handle, mock_notify_bind, mock_conn, mock_get_bind_st,
        mock_save):

        good_states = constants.ARQ_STATES_TRANSFORM_MATRIX[
            constants.ARQ_BIND_STARTED]
        obj_extarq = self.fake_obj_extarqs[0]
        obj_extarq.arq.state = good_states[0]

        # NOTE(Sundar): Since update_check_state is mocked, ARQ state
        # remains as 'Initial'. So, we mock get_arq_bind_statuses to
        # prevent that from raising exception.
        mock_get_bind_st.return_value = [
            (obj_extarq.arq.uuid, constants.ARQ_BIND_STATUS_FINISH)]

        # TODO(Shaohe) we should control the state of arq to make
        # better testcase.
        # bound_extarq = copy.deepcopy(obj_extarq)
        # bound_extarq.arq.state = constants.ARQ_BOUND
        # mock_get.side_effect = [obj_extarq, bound_extarq]

        mock_get.side_effect = [obj_extarq] * 2
        mock_list.return_value = [obj_extarq]
        uuid = obj_extarq.arq.uuid
        instance_uuid = obj_extarq.arq.instance_uuid

        dep_uuid = self.deployable_uuids[0]
        fake_dep = fake_deployable.fake_deployable_obj(self.context,
                                                       uuid=dep_uuid)
        mock_get_dep.return_value = fake_dep
        valid_fields = {
            uuid: {'hostname': obj_extarq.arq.hostname,
                   'device_rp_uuid': obj_extarq.arq.device_rp_uuid,
                   'instance_uuid': instance_uuid}
            }
        patch_list = {
            str(uuid): [
                {"path": "/hostname", "op": "add",
                 "value": obj_extarq.arq.hostname},
                {"path": "/device_rp_uuid", "op": "add",
                 "value": obj_extarq.arq.device_rp_uuid},
                {"path": "/instance_uuid", "op": "add",
                 "value": instance_uuid}
            ]
        }
        objects.ExtARQ.apply_patch(self.context, patch_list, valid_fields)
        # NOTE(Shaohe) we set the fake_obj_extarqs state is ARQ_INITIAL
        # TODO(Shaohe) we should control the state of arq to make
        # complete status testcase.
        self.assertEqual(obj_extarq.arq.state, 'Initial')
        mock_notify_bind.assert_called_once_with(
            instance_uuid,
            [(obj_extarq.arq.uuid, constants.ARQ_BIND_STATUS_FINISH)])

        self.assertEqual(obj_extarq.deployable_id, fake_dep.id)
        mock_save.assert_called_once()

    @mock.patch('cyborg.objects.extarq.ext_arq_job.ExtARQJobMixin.'
                'get_arq_bind_statuses')
    @mock.patch('openstack.connection.Connection')
    @mock.patch('cyborg.common.nova_client.NovaAPI.notify_binding')
    @mock.patch('cyborg.objects.ExtARQ._allocate_attach_handle')
    @mock.patch('cyborg.objects.ExtARQ.get')
    @mock.patch('cyborg.objects.ExtARQ.list')
    @mock.patch('cyborg.objects.ExtARQ.update_check_state')
    @mock.patch('cyborg.objects.deployable.Deployable.get_by_device_rp_uuid')
    @mock.patch('cyborg.common.utils.ThreadWorks.spawn')
    def test_apply_patch_start_fpga_arq_job(
        self, mock_spawn, mock_get_dep, mock_check_state, mock_list, mock_get,
        mock_attach_handle, mock_notify_bind, mock_conn, mock_get_bind_st):
        good_states = constants.ARQ_STATES_TRANSFORM_MATRIX[
            constants.ARQ_BIND_STARTED]
        obj_extarq = self.fake_obj_extarqs[2]
        obj_fpga_extarq = self.fake_obj_fpga_extarqs[1]
        obj_fpga_extarq.state = self.fake_obj_fpga_extarqs[1]
        obj_extarq.arq.state = good_states[0]
        obj_fpga_extarq.arq.state = good_states[0]

        # TODO(Shaohe) we should control the state of arq to make
        # better testcase.
        # bound_extarq = copy.deepcopy(obj_extarq)
        # bound_extarq.arq.state = constants.ARQ_BOUND
        # mock_get.side_effect = [obj_extarq, bound_extarq]
        mock_get.side_effect = [obj_extarq, obj_fpga_extarq]
        mock_list.return_value = [obj_extarq]
        uuid = obj_extarq.arq.uuid
        instance_uuid = obj_extarq.arq.instance_uuid
        # mock_job_get_ext_arq.side_effect = obj_extarq
        dep_uuid = self.deployable_uuids[0]
        fake_dep = fake_deployable.fake_deployable_obj(self.context,
                                                       uuid=dep_uuid)
        mock_get_bind_st.return_value = [
            (obj_extarq.arq.uuid, constants.ARQ_BIND_STATUS_FINISH)]
        mock_get_dep.return_value = fake_dep
        mock_spawn.return_value = None
        valid_fields = {
            uuid: {'hostname': obj_extarq.arq.hostname,
                   'device_rp_uuid': obj_extarq.arq.device_rp_uuid,
                   'instance_uuid': instance_uuid}
            }
        patch_list = {
            str(uuid): [
                {"path": "/hostname", "op": "add",
                 "value": obj_extarq.arq.hostname},
                {"path": "/device_rp_uuid", "op": "add",
                 "value": obj_extarq.arq.device_rp_uuid},
                {"path": "/instance_uuid", "op": "add",
                 "value": instance_uuid}
            ]
        }
        objects.ExtARQ.apply_patch(self.context, patch_list, valid_fields)
        # NOTE(Shaohe) we set the fake_obj_extarqs state is ARQ_INITIAL
        # TODO(Shaohe) we should control the state of arq to make
        # better testcase.
        self.assertEqual(obj_extarq.arq.state, 'Initial')
        mock_notify_bind.assert_called_once_with(
            instance_uuid,
            [(obj_extarq.arq.uuid, constants.ARQ_BIND_STATUS_FINISH)])
        # NOTE(Shaohe) check it spawn to start a job.
        mock_spawn.assert_called_once_with(
            obj_fpga_extarq.bind, self.context, fake_dep)

    @mock.patch('openstack.connection.Connection')
    @mock.patch('cyborg.common.nova_client.NovaAPI.notify_binding')
    @mock.patch('cyborg.objects.ExtARQ._allocate_attach_handle')
    @mock.patch('cyborg.objects.ExtARQ.get')
    @mock.patch('cyborg.objects.ExtARQ.list')
    @mock.patch('cyborg.objects.ExtARQ.update_check_state')
    @mock.patch('cyborg.objects.deployable.Deployable.get_by_device_rp_uuid')
    @mock.patch('cyborg.common.utils.ThreadWorks.spawn_master')
    def test_apply_patch_fpga_arq_monitor_job(
        self, mock_master, mock_get_dep, mock_check_state, mock_list,
        mock_get, mock_attach_handle, mock_notify_bind, mock_conn):

        good_states = constants.ARQ_STATES_TRANSFORM_MATRIX[
            constants.ARQ_BIND_STARTED]
        obj_extarq = self.fake_obj_extarqs[2]
        obj_fpga_extarq = self.fake_obj_fpga_extarqs[1]
        obj_fpga_extarq.state = self.fake_obj_fpga_extarqs[1]
        obj_extarq.arq.state = good_states[0]
        obj_fpga_extarq.arq.state = good_states[0]

        # TODO(Shaohe) we should control the state of arq to make
        # better testcase.
        # bound_extarq = copy.deepcopy(obj_extarq)
        # bound_extarq.arq.state = constants.ARQ_BOUND
        # mock_get.side_effect = [obj_extarq, bound_extarq]
        mock_get.side_effect = [obj_extarq, obj_fpga_extarq]
        mock_list.return_value = [obj_extarq]
        uuid = obj_extarq.arq.uuid
        instance_uuid = obj_extarq.arq.instance_uuid
        dep_uuid = self.deployable_uuids[0]
        fake_dep = fake_deployable.fake_deployable_obj(self.context,
                                                       uuid=dep_uuid)
        mock_get_dep.return_value = fake_dep
        valid_fields = {
            uuid: {'hostname': obj_extarq.arq.hostname,
                   'device_rp_uuid': obj_extarq.arq.device_rp_uuid,
                   'instance_uuid': instance_uuid}
            }
        patch_list = {
            str(uuid): [
                {"path": "/hostname", "op": "add",
                 "value": obj_extarq.arq.hostname},
                {"path": "/device_rp_uuid", "op": "add",
                 "value": obj_extarq.arq.device_rp_uuid},
                {"path": "/instance_uuid", "op": "add",
                 "value": instance_uuid}
            ]
        }
        objects.ExtARQ.apply_patch(self.context, patch_list, valid_fields)
        mock_master.assert_called_once()

    @mock.patch('cyborg.objects.ExtARQ.get')
    @mock.patch('cyborg.objects.ExtARQ._from_db_object')
    def test_destroy(self, mock_from_db_obj, mock_obj_extarq):
        db_extarq = self.fake_db_extarqs[0]
        uuid = db_extarq['uuid']
        mock_from_db_obj.return_value = db_extarq
        mock_obj_extarq.return_value = self.fake_obj_extarqs[0]
        with mock.patch.object(self.dbapi, 'extarq_get',
                               autospec=True) as mock_extarq_get:
            mock_extarq_get.return_value = db_extarq
            with mock.patch.object(self.dbapi, 'extarq_delete',
                                   autospec=True) as mock_extarq_delete:
                extarq = objects.ExtARQ.get(self.context, uuid)
                extarq.destroy(self.context)
                mock_extarq_delete.assert_called_once_with(self.context, uuid)

    @mock.patch('cyborg.objects.ExtARQ.update_check_state')
    def test_allocate_attach_handle(self, mock_check_state):
        obj_extarq = self.fake_obj_extarqs[0]
        dep_uuid = self.deployable_uuids[0]
        fake_dep = fake_deployable.fake_deployable_obj(self.context,
                                                       uuid=dep_uuid)
        self.assertRaises(
            exception.ResourceNotFound,
            obj_extarq._allocate_attach_handle, self.context, fake_dep)
        mock_check_state.assert_called_once_with(
            self.context, constants.ARQ_BIND_FAILED)

    @mock.patch('logging.LoggerAdapter.error')
    @mock.patch('cyborg.objects.attach_handle.AttachHandle.allocate')
    @mock.patch('cyborg.objects.ExtARQ.update_check_state')
    def test_allocate_attach_handle_with_error_log(
        self, mock_check_state, mock_allocate, mock_log):
        obj_extarq = self.fake_obj_extarqs[0]
        dep_uuid = self.deployable_uuids[0]
        e = exception.ResourceNotFound(
            resource='AttachHandle', msg="Just for Test")
        msg = ("Failed to allocate attach handle for ARQ %s"
               "from deployable %s. Reason: %s")
        mock_allocate.side_effect = e
        fake_dep = fake_deployable.fake_deployable_obj(self.context,
                                                       uuid=dep_uuid)
        self.assertRaises(
            exception.ResourceNotFound,
            obj_extarq._allocate_attach_handle, self.context, fake_dep)
        mock_log.assert_called_once_with(
            msg, obj_extarq.arq.uuid, fake_dep.uuid, str(e))
        mock_check_state.assert_called_once_with(
            self.context, constants.ARQ_BIND_FAILED)

    @mock.patch('cyborg.objects.ExtARQ.update_check_state')
    @mock.patch('cyborg.objects.attach_handle.AttachHandle.get_by_id')
    @mock.patch('cyborg.objects.attach_handle.AttachHandle.deallocate')
    def test_deallocate_attach_handle(
        self, mock_deallocate, mock_ah, mock_check_state):
        obj_extarq = self.fake_obj_extarqs[0]
        mock_ah.return_value = self.fake_obj_ahs[0]
        obj_extarq._deallocate_attach_handle(
            self.context, mock_ah.id, obj_extarq.arq.hostname)
        mock_check_state.assert_not_called()

    @mock.patch('logging.LoggerAdapter.error')
    @mock.patch('cyborg.objects.ExtARQ.update_check_state')
    @mock.patch('cyborg.objects.attach_handle.AttachHandle.get_by_id')
    @mock.patch('cyborg.objects.attach_handle.AttachHandle.deallocate')
    def test_deallocate_attach_handle_with_error_log(
        self, mock_ah, mock_deallocate, mock_check_state, mock_log):
        obj_extarq = self.fake_obj_extarqs[0]
        mock_ah.return_value = self.fake_obj_ahs[0]
        e = exception.ResourceNotFound(
            resource='AttachHandle', msg="Just for Test")
        msg = ("Failed to deallocate attach handle %s for ARQ %s."
               "Reason: %s")
        mock_deallocate.side_effect = e
        self.assertRaises(
            exception.ResourceNotFound,
            obj_extarq._deallocate_attach_handle, self.context, mock_ah.id,
            obj_extarq.arq.hostname)
        mock_log.assert_called_once_with(
            msg, mock_ah.id, obj_extarq.arq.uuid, str(e))
        mock_check_state.assert_called_once_with(
            self.context, constants.ARQ_UNBIND_FAILED)

    @mock.patch('cyborg.objects.ExtARQ.get')
    @mock.patch('cyborg.objects.ExtARQ._from_db_object')
    def test_save(self, mock_from_db_obj, mock_obj_extarq):
        db_extarq = self.fake_db_extarqs[0]
        uuid = db_extarq['uuid']
        mock_from_db_obj.return_value = db_extarq
        mock_obj_extarq.return_value = self.fake_obj_extarqs[0]
        with mock.patch.object(self.dbapi, 'extarq_update',
                               autospec=True) as mock_extarq_update:
            obj_extarq = objects.ExtARQ.get(self.context, uuid)
            obj_extarq.arq.hostname = 'newtestnode1'
            fake_arq_updated = db_extarq
            fake_arq_updated['hostname'] = obj_extarq.arq.hostname
            mock_extarq_update.return_value = fake_arq_updated
            obj_extarq.save(self.context)
            mock_extarq_update.assert_called_once()

    def test_get_arq_bind_statuses(self):
        # ARQ state is 'Bound'  by default in the fake extarqs
        arq_list = [extarq.arq for extarq in self.fake_obj_extarqs]
        bind_status = constants.ARQ_BIND_STATES_STATUS_MAP[
            constants.ARQ_BOUND]
        expected = [(arq.uuid, bind_status) for arq in arq_list]

        result = objects.ExtARQ.get_arq_bind_statuses(arq_list)

        self.assertEqual(expected, result)

    def _test_get_arq_bind_statuses_exception(self):
        extarqs = fake_extarq.get_fake_extarq_objs()
        arq_list = [extarq.arq for extarq in extarqs]
        for arq in arq_list:
            arq['state'] = constants.ARQ_INITIAL

        self.assertRaises(exception.ARQBadState,
                          objects.ExtARQ.get_arq_bind_statuses, arq_list)

    @mock.patch('cyborg.objects.device_profile.DeviceProfile.get_by_name')
    @mock.patch('cyborg.objects.deployable.Deployable.get_by_id')
    @mock.patch('cyborg.db.sqlalchemy.api.Connection.'
                'attach_handle_get_by_id')
    @mock.patch('cyborg.db.sqlalchemy.api.Connection.'
                'device_profile_get_by_id')
    def test_fill_obj_extarq_fields(self, mock_get_devprof, mock_get_ah,
                                    mock_get_obj_dep, mock_obj_devprof):
        in_db_extarq = self.fake_db_extarqs[0]
        # Since state is not 'Bound', attach_handle_get_by_id is not called.
        in_db_extarq['state'] = 'Initial'
        in_db_extarq['deployable_id'] = None
        db_devprof = fake_device_profile.get_db_devprofs()[0]
        obj_devprof = fake_device_profile.get_obj_devprofs()[0]

        mock_get_devprof.return_value = db_devprof
        mock_obj_devprof.return_value = obj_devprof

        out_db_extarq = objects.ExtARQ._fill_obj_extarq_fields(
            self.context, in_db_extarq)

        self.assertEqual(out_db_extarq['device_profile_name'],
                         db_devprof['name'])
        self.assertEqual(out_db_extarq['attach_handle_type'], '')
        self.assertEqual(out_db_extarq['attach_handle_info'], '')
        self.assertEqual(
            out_db_extarq['deployable_uuid'],
            '00000000-0000-0000-0000-000000000000')
        devprof_group_id = out_db_extarq['device_profile_group_id']
        self.assertEqual(out_db_extarq['device_profile_group'],
                         obj_devprof['groups'][devprof_group_id])

    @mock.patch('cyborg.objects.device_profile.DeviceProfile.get_by_name')
    @mock.patch('cyborg.objects.deployable.Deployable.get_by_id')
    @mock.patch('cyborg.db.sqlalchemy.api.Connection.'
                'attach_handle_get_by_id')
    @mock.patch('cyborg.db.sqlalchemy.api.Connection.'
                'device_profile_get_by_id')
    def test_fill_obj_extarq_fields_with_dep_id(
            self, mock_get_devprof, mock_get_ah, mock_get_obj_dep,
            mock_obj_devprof):
        in_db_extarq = self.fake_db_extarqs[0]
        # Since state is not 'Bound', attach_handle_get_by_id is not called.
        in_db_extarq['state'] = 'Initial'
        in_db_extarq['deployable_id'] = '1'
        db_devprof = fake_device_profile.get_db_devprofs()[0]
        obj_devprof = fake_device_profile.get_obj_devprofs()[0]

        mock_get_devprof.return_value = db_devprof
        mock_obj_devprof.return_value = obj_devprof
        mock_get_obj_dep.return_value = \
            fake_deployable.fake_deployable_obj(self.context,
                                                uuid=self.deployable_uuids[0])

        out_db_extarq = objects.ExtARQ._fill_obj_extarq_fields(
            self.context, in_db_extarq)

        self.assertEqual(out_db_extarq['device_profile_name'],
                         db_devprof['name'])
        self.assertEqual(out_db_extarq['attach_handle_type'], '')
        self.assertEqual(out_db_extarq['attach_handle_info'], '')
        devprof_group_id = out_db_extarq['device_profile_group_id']
        self.assertEqual(out_db_extarq['device_profile_group'],
                         obj_devprof['groups'][devprof_group_id])

    def test_obj_make_compatible(self):
        arq_obj = objects.ExtARQ(deployable_id=1)
        primitive = arq_obj.obj_to_primitive()
        arq_obj.obj_make_compatible(primitive['cyborg_object.data'], '1.1')
        self.assertNotIn('deployable_id', primitive['cyborg_object.data'])
        primitive = arq_obj.obj_to_primitive()
        arq_obj.obj_make_compatible(primitive['cyborg_object.data'], '1.2')
        self.assertIn('deployable_id', primitive['cyborg_object.data'])



# Copyright 2019 Intel, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import HTTPStatus
import pecan
import wsme
from wsme import types as wtypes

from oslo_log import log

from cyborg.accelerator.common import exception
from cyborg import api
from cyborg.api.controllers import base
from cyborg.api.controllers import link
from cyborg.api.controllers import types
from cyborg.api.controllers.v2 import versions
from cyborg.api import expose
from cyborg.common import authorize_wsgi
from cyborg.common import placement_client
from cyborg import objects


LOG = log.getLogger(__name__)


class Device(base.APIBase):
    """API representation of a device.

    This class enforces type checking and value constraints, and converts
    between the internal object model and the API representation.
    """
    uuid = types.uuid
    """The UUID of the device"""

    type = wtypes.text
    """The type of the device"""

    vendor = wtypes.text
    """The vendor of the device"""

    model = wtypes.text
    """The model of the device"""

    std_board_info = wtypes.text
    """The standard board info of the device"""

    vendor_board_info = wtypes.text
    """The vendor board info of the device"""

    hostname = wtypes.text
    """The hostname of the device"""

    status = wtypes.text
    """The status of the device"""

    links = wsme.wsattr([link.Link], readonly=True)
    """A list containing a self link"""

    def __init__(self, **kwargs):
        super(Device, self).__init__(**kwargs)
        self.fields = []
        for field in objects.Device.fields:
            self.fields.append(field)
            setattr(self, field, kwargs.get(field, wtypes.Unset))

    @classmethod
    def convert_with_links(cls, obj_device):
        api_device = cls(**obj_device.as_dict())
        # return status filed from microversion of 2.3
        if not api.request.version.minor >= versions.MINOR_3_DEVICE_STATUS:
            delattr(api_device, 'status')
        api_device.links = [
            link.Link.make_link('self', pecan.request.public_url,
                                'devices', api_device.uuid)
            ]
        return api_device


class DeviceCollection(base.APIBase):
    """API representation of a collection of devices."""

    devices = [Device]
    """A list containing Device objects"""

    @classmethod
    def convert_with_links(cls, devices):
        collection = cls()
        collection.devices = [Device.convert_with_links(device)
                              for device in devices]
        return collection


class DevicesController(base.CyborgController):
    """REST controller for Devices."""

    _custom_actions = {'disable': ['POST'], 'enable': ['POST']}

    @authorize_wsgi.authorize_wsgi("cyborg:device", "get_one")
    @expose.expose(Device, wtypes.text)
    def get_one(self, uuid):
        """Get a single device by UUID.
        :param uuid: uuid of a device.
        """
        context = pecan.request.context
        device = objects.Device.get(context, uuid)
        return Device.convert_with_links(device)

    @authorize_wsgi.authorize_wsgi("cyborg:device", "get_all", False)
    @expose.expose(DeviceCollection, wtypes.text, wtypes.text, wtypes.text,
                   wtypes.ArrayType(types.FilterType))
    def get_all(self, type=None, vendor=None, hostname=None, filters=None):
        """Retrieve a list of devices.
        :param type: type of a device.
        :param vendor: vendor ID of a device.
        :param hostname: the hostname of a compute node where the device
        locates.
        :param filters: a filter of FilterType to get device list by filter.
        """
        filters_dict = {}
        if type:
            filters_dict["type"] = type
        if vendor:
            filters_dict["vendor"] = vendor
        if hostname:
            filters_dict["hostname"] = hostname
        if filters:
            for filter in filters:
                filters_dict.update(filter.as_dict())
        context = pecan.request.context
        obj_devices = objects.Device.list(context, filters=filters_dict)
        LOG.info('[devices:get_all] Returned: %s', obj_devices)
        return DeviceCollection.convert_with_links(obj_devices)

    @authorize_wsgi.authorize_wsgi("cyborg:device", "disable")
    @expose.expose(None, wtypes.text, types.uuid,
                   status_code=HTTPStatus.OK)
    def disable(self, uuid):
        context = pecan.request.context
        device = objects.Device.get(context, uuid)
        device.status = 'maintaining'
        device.save(context)
        # update resource provider inventories
        client = placement_client.PlacementClient()
        deployable = objects.Deployable.get_by_id(context, device.id)
        filters = {'deployable_id': deployable.id, 'key': 'rc'}
        attributes = objects.Attribute.get_by_filter(context, filters)
        if attributes:
            att_type = attributes[0].value
        else:
            raise exception.ResourceNotFound(
                resource='Attribute',
                msg='with deployable_id=%s,key=%s' % (deployable.id, 'rc'))
        client.update_rp_inventory_reserved(
            deployable.rp_uuid, att_type,
            deployable.num_accelerators,
            deployable.num_accelerators)

    @authorize_wsgi.authorize_wsgi("cyborg:device", "enable")
    @expose.expose(None, wtypes.text, types.uuid,
                   status_code=HTTPStatus.OK)
    def enable(self, uuid):
        context = pecan.request.context
        device = objects.Device.get(context, uuid)
        device.status = 'enabled'
        device.save(context)
        # update resource provider inventories
        client = placement_client.PlacementClient()
        deployable = objects.Deployable.get_by_id(context, device.id)
        filters = {'deployable_id': deployable.id, 'key': 'rc'}
        attributes = objects.Attribute.get_by_filter(context, filters)
        if attributes:
            att_type = attributes[0].value
        else:
            raise exception.ResourceNotFound(
                resource='Attribute',
                msg='with deployable_id=%s,key=%s' % (deployable.id, 'rc'))
        client.update_rp_inventory_reserved(
            deployable.rp_uuid, att_type,
            deployable.num_accelerators,
            0)



# Copyright 2020 Intel, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
Cyborg Intel QAT driver implementation.
"""


import glob
import os
import socket

from cyborg.accelerator.common import utils
from cyborg.common import constants
from cyborg.objects.driver_objects import driver_attach_handle
from cyborg.objects.driver_objects import driver_attribute
from cyborg.objects.driver_objects import driver_controlpath_id
from cyborg.objects.driver_objects import driver_deployable
from cyborg.objects.driver_objects import driver_device

from oslo_serialization import jsonutils


PCI_DEVICES_PATH = "/sys/bus/pci/devices"
KNOW_QATS = [("0x8086", "0x37c8")]
PF = "physfn"
VF = "virtfn*"
INTEL_QAT_DEV_PREFIX = "intel-qat-dev"
RC_QAT = constants.RESOURCES["QAT"]
DRIVER_NAME = "intel"

RESOURCES = {
    "qat": RC_QAT
}


def pci_attributes(path):
    with open(os.path.join(path, "uevent")) as f:
        attributes = dict(map(
            lambda p: p.strip().split("="),
            f.readlines()
        ))

    with open(os.path.join(path, "vendor")) as f:
        attributes["VENDOR"] = f.readline().strip()

    with open(os.path.join(path, "device")) as f:
        attributes["PRODUCT_ID"] = f.readline().strip()

    return attributes


def get_link_targets(links):
    return map(
        lambda p:
            os.path.realpath(
                os.path.join(os.path.dirname(p), os.readlink(p))),
        links)


def all_qats():
    return set(filter(
        lambda p: (
            pci_attributes(p)["VENDOR"],
            pci_attributes(p)["PRODUCT_ID"]
        ) in KNOW_QATS,
        glob.glob(os.path.join(PCI_DEVICES_PATH, "*"))))


def all_pfs_with_vf():
    return set(filter(
        lambda p: glob.glob(os.path.join(p, VF)),
        all_qats()))


def all_vfs_in_pf(pf_path):
    return map(
        lambda p:
        os.path.join(
            os.path.dirname(os.path.dirname(p)),
            os.path.basename(os.readlink(p))),
        glob.glob(os.path.join(pf_path, VF)))


def find_pf_by_vf(vf_path):
    return os.path.join(
        os.path.dirname(vf_path),
        os.path.basename(os.readlink(
            os.path.join(vf_path, PF))))


def all_vfs():
    return map(
        lambda p: all_vfs_in_pf(p), all_pfs_with_vf()
    )


def qat_gen(path):
    pci_info = pci_attributes(path)
    qat = {
        "name": "_".join((socket.gethostname(), pci_info["PCI_SLOT_NAME"])),
        "device": pci_info["PCI_SLOT_NAME"],
        "type": constants.DEVICE_QAT,
        "vendor": pci_info["VENDOR"],
        "product_id": pci_info["PRODUCT_ID"],
        "rc": RESOURCES["qat"],
        "stub": False,
    }
    return qat


def qat_tree():
    devs = []
    pfs_has_vf = all_pfs_with_vf()
    for q in all_qats():
        qat = qat_gen(q)
        if q in pfs_has_vf:
            vfs = []
            for vf in all_vfs_in_pf(q):
                vf_qat = qat_gen(vf)
                vfs.append(vf_qat)
            qat["vfs"] = vfs
        devs.append(_generate_driver_device(qat))
    return devs


def _generate_driver_device(qat):
    driver_device_obj = driver_device.DriverDevice()
    driver_device_obj.vendor = qat["vendor"]
    driver_device_obj.stub = qat["stub"]
    driver_device_obj.model = qat.get("model", "miss_model_info")
    driver_device_obj.vendor_board_info = qat.get(
        "vendor_board_info",
        "miss_vb_info")
    std_board_info = {"product_id": qat.get("product_id")}
    driver_device_obj.std_board_info = jsonutils.dumps(std_board_info)
    driver_device_obj.type = qat["type"]
    driver_device_obj.controlpath_id = _generate_controlpath_id(qat)
    driver_device_obj.deployable_list = _generate_dep_list(qat)
    return driver_device_obj


def _generate_controlpath_id(qat):
    driver_cpid = driver_controlpath_id.DriverControlPathID()
    driver_cpid.cpid_type = "PCI"
    driver_cpid.cpid_info = utils.pci_str_to_json(qat["device"])
    return driver_cpid


def _generate_dep_list(qat):
    dep_list = []
    # pf without sriov enabled.
    if "vfs" not in qat:
        driver_dep = driver_deployable.DriverDeployable()
        driver_dep.num_accelerators = 1
        driver_dep.attach_handle_list = [
            _generate_attach_handle(qat)]
        driver_dep.name = qat["name"]
        driver_dep.driver_name = DRIVER_NAME
        driver_dep.attribute_list = _generate_attribute_list(qat)
        dep_list = [driver_dep]
    # pf with sriov enabled, may have several vfs.
    else:
        for vf in qat["vfs"]:
            driver_dep = driver_deployable.DriverDeployable()
            driver_dep.num_accelerators = 1
            driver_dep.attach_handle_list = [
                _generate_attach_handle(vf)]
            driver_dep.name = vf["name"]
            driver_dep.driver_name = DRIVER_NAME
            driver_dep.attribute_list = _generate_attribute_list(qat)
            dep_list.append(driver_dep)
    return dep_list


def _generate_attach_handle(qat):
    driver_ah = driver_attach_handle.DriverAttachHandle()
    driver_ah.attach_type = "PCI"
    driver_ah.attach_info = utils.pci_str_to_json(qat["device"])
    driver_ah.in_use = False
    return driver_ah


def _generate_attribute_list(qat):
    attr_list = []
    for k, v in qat.items():
        if k == "rc":
            driver_attr = driver_attribute.DriverAttribute()
            driver_attr.key = k
            driver_attr.value = v
            attr_list.append(driver_attr)
    return attr_list



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import importlib
import pecan
from pecan import rest
from wsme import types as wtypes

from cyborg.api.controllers import base
from cyborg.api.controllers import link
from cyborg.api.controllers import v2
from cyborg.api import expose


class APIStatus(object):
    CURRENT = "CURRENT"
    SUPPORTED = "SUPPORTED"
    DEPRECATED = "DEPRECATED"
    EXPERIMENTAL = "EXPERIMENTAL"


class Version(base.APIBase):
    """An API version representation."""

    id = wtypes.text
    """The ID of the version, also acts as the release number"""

    status = wtypes.text
    """The state of this API version"""

    max_version = wtypes.text
    """The maximum version supported"""

    min_version = wtypes.text
    """The minimum version supported"""

    links = [link.Link]
    """A Link that points to a specific version of the API"""

    @staticmethod
    def convert(id, status=APIStatus.CURRENT):
        version = Version()
        if id == "v1":
            version.max_version = None
            version.min_version = None
        else:
            v = importlib.import_module(
                'cyborg.api.controllers.%s.versions' % id)
            version.max_version = v.max_version_string()
            version.min_version = v.min_version_string()
        version.id = id
        version.status = status
        version.links = [link.Link.make_link('self', pecan.request.host_url,
                                             id, '', bookmark=True)]
        return version


class Root(base.APIBase):
    name = wtypes.text
    """The name of the API"""

    description = wtypes.text
    """Some information about this API"""

    versions = [Version]
    """Links to all the versions available in this API"""

    default_version = Version
    """A link to the default version of the API"""

    @staticmethod
    def convert():
        root = Root()
        root.name = 'OpenStack Cyborg API'
        root.description = (
            "Cyborg is the OpenStack project for lifecycle "
            "management of hardware accelerators, such as GPUs,"
            "FPGAs, AI chips, security accelerators, etc.")
        root.versions = [Version.convert('v2')]
        root.default_version = Version.convert('v2')
        return root


class RootController(rest.RestController):
    _versions = [base.API_V2]
    """All supported API versions"""

    _default_version = base.API_V2
    """The default API version"""

    v2 = v2.Controller()

    @expose.expose(Root)
    def get(self):
        return Root.convert()

    @pecan.expose()
    def _route(self, args, request=None):
        """Overrides the default routing behavior.

        It redirects the request to the default version of the cyborg API
        if the version number is not specified in the url.
        """

        if args[0] and args[0] not in self._versions:
            args = [self._default_version] + args
        return super(RootController, self)._route(args)



# Copyright 2019 Intel Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from oslo_serialization import jsonutils

from cyborg.objects import device_profile

"""
   See note at the start of cyborg/api/controllers/v2/device_profiles.py.
   Device profiles have an API format (which is provided to POST to
   create one) and an object format. The code in this file can provide
   fake device profiles in either format.
"""


def _get_device_profiles_as_dict():
    date1 = datetime.datetime(
        2019, 10, 9, 6, 31, 59,
        tzinfo=datetime.timezone.utc)
    date2 = datetime.datetime(
        2019, 11, 8, 5, 30, 49,
        tzinfo=datetime.timezone.utc)
    dp1 = {
        "id": 1,
        "uuid": "a95e10ae-b3e3-4eab-a513-1afae6f17c51",
        "name": 'dp_example_1',
        "description": "fake-dp_example_1-desc",
        "created_at": date1,
        "updated_at": None,
        "groups": [
            {"resources:FPGA": "1",
             "trait:CUSTOM_FPGA_INTEL_PAC_ARRIA10": "required",
             "trait:CUSTOM_FUNCTION_ID_3AFB": "required",
             },
            {"resources:CUSTOM_ACCELERATOR_FOO": "2",
             "trait:CUSTOM_TRAIT_ALWAYS": "required",
             }
        ]
    }
    dp2 = {
        "id": 2,
        "uuid": u"199c46b7-63a7-431b-aa40-35da4b9420b1",
        "name": 'dp_example_2',
        "created_at": date2,
        "updated_at": None,
        "description": "fake-dp_example_2-desc",
        "groups": [
            {"resources:FPGA": "1",
             "trait:CUSTOM_REGION_ID_3ACD": "required",
             "accel:bitstream_id": "ea0d149c-8555-495b-bc79-608d7bab1260"
             }
        ]
    }
    return [dp1, dp2]


def _convert_to_obj(dp_dict):
    obj_devprof = device_profile.DeviceProfile()
    for field in dp_dict.keys():
        obj_devprof[field] = dp_dict[field]
    return obj_devprof


def _convert_to_db_devprof(dp_dict):
    profile_json_dict = {'groups': dp_dict['groups']}
    profile_json = jsonutils.dumps(profile_json_dict)
    dp_dict['profile_json'] = profile_json
    del dp_dict['groups']
    return dp_dict


def _drop_uuid(dp_dict):
    del dp_dict['uuid']
    return dp_dict


def get_obj_devprofs():
    dp_list = _get_device_profiles_as_dict()
    obj_devprofs = list(map(_convert_to_obj, dp_list))
    return obj_devprofs


def get_api_devprofs(drop_uuid=False):
    dp_list = _get_device_profiles_as_dict()
    if drop_uuid:
        api_devprofs = list(map(_drop_uuid, dp_list))
    else:
        api_devprofs = dp_list
    return api_devprofs


def get_db_devprofs():
    dp_list = _get_device_profiles_as_dict()
    db_devprofs = list(map(_convert_to_db_devprof, dp_list))
    return db_devprofs


def get_xilinx_fpga_devprof():
    xilinx_fpga_dp = {
        "id": 3,
        "uuid": "638828d7-7145-4060-904c-894aa726b133",
        "name": 'fake_xilinx_fpga_dp',
        "description": "fake_xilinx_fpga_dp-desc",
        "created_at": datetime.datetime(
            2022, 1, 22, 10, 40, 56,
            tzinfo=datetime.timezone.utc),
        "updated_at": None,
        "groups": [
            {"resources:FPGA": "1",
             "trait:CUSTOM_FPGA_XILINX": "required",
             "trait:CUSTOM_FPGA_PRODUCT_ID_5000": "required",
             },
        ]
    }
    return _convert_to_obj(xilinx_fpga_dp)



# Copyright 2020 Intel, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
Cyborg Intel NIC driver implementation.
"""


import glob
import os
import socket

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils

from cyborg.accelerator.common import utils
import cyborg.conf

from cyborg.objects.driver_objects import driver_attach_handle
from cyborg.objects.driver_objects import driver_attribute
from cyborg.objects.driver_objects import driver_controlpath_id
from cyborg.objects.driver_objects import driver_deployable
from cyborg.objects.driver_objects import driver_device

LOG = logging.getLogger(__name__)

PCI_DEVICES_PATH_PATTERN = "/sys/bus/pci/devices/*"
KNOWN_NICS = [("0x8086", "0x158b"), ("0x8086", "0x1572")]
DRIVER_NAME = "intel"

VF = "virtfn*"
_SRIOV_TOTALVFS = "sriov_totalvfs"
CONF = cyborg.conf.CONF


def _parse_config():
    # parse nic config.
    cyborg.conf.devices.register_dynamic_opts(CONF)
    try:
        pdm = utils.parse_mappings(CONF.x710_static.physical_device_mappings)
        fdm = utils.parse_mappings(CONF.x710_static.function_device_mappings)
    except cfg.NoSuchOptError:
        return None, None
    else:
        return pdm, fdm


def get_physical_network_and_traits(pci_info, physnet_device_mappings,
                                    function_device_mappings, pf_nic=None):
    traits = []
    physnet = None
    func_name = None

    if pf_nic:
        pf_addr = pf_nic["device"]
        traits.append("CUSTOM_VF")
    else:
        pf_addr = pci_info["PCI_SLOT_NAME"]
        traits.append("CUSTOM_PF")
    if not pf_addr:
        LOG.error("Incorrect report data received. Missing PF info.")

    pf_ifname = utils.get_ifname_by_pci_address(pf_addr)

    if physnet_device_mappings:
        for key, devices in physnet_device_mappings.items():
            if pf_ifname in devices:
                physnet = key
                break
    if function_device_mappings:
        for key, devices in function_device_mappings.items():
            if pf_ifname in devices:
                func_name = key
                break
    if func_name:
        traits.append("CUSTOM_" + func_name.upper())
    if physnet:
        traits.append("CUSTOM_" + physnet.upper())

    return {"traits": traits, "physical_network": physnet}


def read_line(filename):
    with open(filename) as f:
        return f.readline().strip()


def find_nics_by_know_list():
    return set(filter(
        lambda p: (
            read_line(os.path.join(p, "vendor")),
            read_line(os.path.join(p, "device"))
        ) in KNOWN_NICS,
        glob.glob(PCI_DEVICES_PATH_PATTERN)))


def pci_attributes(path):
    with open(os.path.join(path, "uevent")) as f:
        attributes = dict(map(
            lambda p: p.strip().split("="),
            f.readlines()
        ))

    with open(os.path.join(path, "vendor")) as f:
        attributes["VENDOR"] = f.readline().strip()

    with open(os.path.join(path, "device")) as f:
        attributes["PRODUCT_ID"] = f.readline().strip()

    return attributes


def nic_gen(path, physnet_device_mappings=None, function_device_mappings=None,
            pf_nic=None):
    pci_info = pci_attributes(path)
    nic = {
        "name": "_".join((socket.gethostname(), pci_info["PCI_SLOT_NAME"])),
        "device": pci_info["PCI_SLOT_NAME"],
        "type": "NIC",
        "vendor": pci_info["VENDOR"],
        "product_id": pci_info["PRODUCT_ID"],
        "rc": "CUSTOM_NIC",
        "stub": False,
        }
    # TODO(Xinran): need check device id and call get_traits differently.
    updates = get_physical_network_and_traits(pci_info,
                                              physnet_device_mappings,
                                              function_device_mappings,
                                              pf_nic)

    nic.update(updates)
    return nic


def all_pfs_with_vf():
    return set(filter(
        lambda p: glob.glob(os.path.join(p, VF)),
        find_nics_by_know_list()))


def all_vfs_in_pf(pf_path):
    return map(
        lambda p:
        os.path.join(
            os.path.dirname(os.path.dirname(p)),
            os.path.basename(os.readlink(p))),
        glob.glob(os.path.join(pf_path, VF)))


def nic_tree():
    physnet_device_mappings, function_device_mappings = _parse_config()
    nics = []
    pfs_has_vf = all_pfs_with_vf()
    for n in find_nics_by_know_list():
        nic = nic_gen(n, physnet_device_mappings, function_device_mappings)
        if n in pfs_has_vf:
            vfs = []
            for vf in all_vfs_in_pf(n):
                vf_nic = nic_gen(vf, physnet_device_mappings,
                                 function_device_mappings, nic)
                vfs.append(vf_nic)
            nic["vfs"] = vfs
        nics.append(_generate_driver_device(nic))
    return nics


def _generate_driver_device(nic):
    driver_device_obj = driver_device.DriverDevice()
    driver_device_obj.vendor = nic["vendor"]
    driver_device_obj.stub = nic["stub"]
    driver_device_obj.model = nic.get("model", "miss_model_info")
    driver_device_obj.vendor_board_info = nic.get(
        "vendor_board_info",
        "miss_vb_info")
    std_board_info = {"product_id": nic.get("product_id")}
    driver_device_obj.std_board_info = jsonutils.dumps(std_board_info)
    driver_device_obj.type = nic["type"]
    driver_device_obj.controlpath_id = _generate_controlpath_id(nic)
    driver_device_obj.deployable_list = _generate_dep_list(nic)
    return driver_device_obj


def _generate_controlpath_id(nic):
    driver_cpid = driver_controlpath_id.DriverControlPathID()
    driver_cpid.cpid_type = "PCI"
    driver_cpid.cpid_info = utils.pci_str_to_json(nic["device"])
    return driver_cpid


def _generate_dep_list(nic):
    dep_list = []
    # pf without sriov enabled.
    if "vfs" not in nic:
        driver_dep = driver_deployable.DriverDeployable()
        driver_dep.num_accelerators = 1
        driver_dep.attach_handle_list = [
            _generate_attach_handle(nic)]
        driver_dep.name = nic["name"]
        driver_dep.driver_name = DRIVER_NAME
        driver_dep.attribute_list = _generate_attribute_list(nic)
        dep_list = [driver_dep]
    # pf with sriov enabled, may have several vfs.
    else:
        for vf in nic["vfs"]:
            driver_dep = driver_deployable.DriverDeployable()
            driver_dep.num_accelerators = 1
            driver_dep.attach_handle_list = [
                _generate_attach_handle(vf)]
            driver_dep.name = vf["name"]
            driver_dep.driver_name = DRIVER_NAME
            driver_dep.attribute_list = _generate_attribute_list(vf)
            dep_list.append(driver_dep)
    return dep_list


def _generate_attach_handle(nic):
    driver_ah = driver_attach_handle.DriverAttachHandle()
    driver_ah.attach_type = "PCI"
    driver_ah.attach_info = utils.pci_str_to_json(nic["device"],
                                                  nic["physical_network"])
    driver_ah.in_use = False
    return driver_ah


def _generate_attribute_list(nic):
    attr_list = []
    for k, v in nic.items():
        if k == "rc":
            driver_attr = driver_attribute.DriverAttribute()
            driver_attr.key, driver_attr.value = k, v
            attr_list.append(driver_attr)
        if k == "traits":
            values = nic.get(k, [])
            for index, val in enumerate(values):
                driver_attr = driver_attribute.DriverAttribute()
                driver_attr.key = "trait" + str(index)
                driver_attr.value = val
                attr_list.append(driver_attr)

    return attr_list



# Copyright 2020 ZTE Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for manipulating Deployable via the DB API"""

from oslo_utils import uuidutils

from cyborg.common import exception
from cyborg.tests.unit.db import base
from cyborg.tests.unit.db import utils


class TestDbDeployable(base.DbTestCase):

    def test_create(self):
        kw = {'name': 'test_create_dep'}
        created_dep = utils.create_test_deployable(self.context, **kw)
        self.assertEqual(created_dep['name'], kw['name'])

    def test_get_by_uuid(self):
        created_dep = utils.create_test_deployable(self.context)
        queried_dep = self.dbapi.deployable_get(
            self.context, created_dep['uuid'])
        self.assertEqual(created_dep['uuid'], queried_dep['uuid'])

    def test_get_by_rp_uuid(self):
        created_dep = utils.create_test_deployable(self.context)
        queried_dep = self.dbapi.deployable_get_by_rp_uuid(
            self.context, created_dep['rp_uuid'])
        self.assertEqual(created_dep['uuid'], queried_dep['uuid'])

    def test_update(self):
        created_dep = utils.create_test_deployable(self.context)
        bit_stream_id = '10efe63d-dfea-4a37-ad94-4116fba5011'
        queried_dep = self.dbapi.deployable_update(
            self.context, created_dep['uuid'],
            {'bit_stream_id': bit_stream_id})
        self.assertEqual(bit_stream_id, queried_dep['bit_stream_id'])

    def test_list(self):
        uuids = []
        for i in range(1, 4):
            dep = utils.create_test_deployable(
                self.context,
                id=i,
                uuid=uuidutils.generate_uuid())
            uuids.append(dep['uuid'])
        deps = self.dbapi.deployable_list(self.context)
        dep_uuids = [item.uuid for item in deps]
        self.assertEqual(sorted(uuids), sorted(dep_uuids))

    def test_delete(self):
        created_dep = utils.create_test_deployable(self.context)
        return_value = self.dbapi.deployable_delete(
            self.context,
            created_dep['uuid'])
        self.assertIsNone(return_value)

    def test_list_by_filters(self):
        dep1 = utils.create_test_deployable(
            self.context,
            id=1,
            uuid=uuidutils.generate_uuid(),
            name='mydep1')
        utils.create_test_deployable(
            self.context,
            id=2,
            uuid=uuidutils.generate_uuid(),
            name='mydep2')
        res = self.dbapi.deployable_get_by_filters(
            self.context, filters={"name": "mydep1"})
        self.assertEqual(1, len(res))
        self.assertEqual(dep1['name'], res[0]['name'])

    def test_list_filter_is_none(self):
        """The main test is filters=None. If filters=None,
        it will be initialized to {}, that will return all deployable
        same as the List Deployable API response.
        """
        dep1 = utils.create_test_deployable(
            self.context,
            id=1,
            uuid=uuidutils.generate_uuid())
        res = self.dbapi.deployable_get_by_filters(
            self.context, filters=None)
        self.assertEqual(1, len(res))
        self.assertEqual(dep1['uuid'], res[0]['uuid'])

    def test_get_by_uuid_not_exist(self):
        random_uuid = uuidutils.generate_uuid()
        self.assertRaises(exception.ResourceNotFound,
                          self.dbapi.deployable_get,
                          self.context, random_uuid)

    def test_get_by_rp_uuid_not_exist(self):
        random_uuid = uuidutils.generate_uuid()
        self.assertRaises(exception.ResourceNotFound,
                          self.dbapi.deployable_get_by_rp_uuid,
                          self.context, random_uuid)

    def test_delete_by_uuid_not_exist(self):
        random_uuid = uuidutils.generate_uuid()
        self.assertRaises(exception.ResourceNotFound,
                          self.dbapi.deployable_delete,
                          self.context, random_uuid)



# -*- encoding: utf-8 -*-
# Copyright (c) 2019 ZTE Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import versionutils
from oslo_versionedobjects import base as object_base

from cyborg.common import exception
from cyborg.db import api as dbapi
from cyborg.objects import base
from cyborg.objects import fields as object_fields


LOG = logging.getLogger(__name__)


@base.CyborgObjectRegistry.register
class DeviceProfile(base.CyborgObject, object_base.VersionedObjectDictCompat):
    # Version 1.0: Initial version
    # Version 1.1: Added description field.
    VERSION = '1.1'

    dbapi = dbapi.get_instance()

    fields = {
        'id': object_fields.IntegerField(nullable=False),
        'uuid': object_fields.StringField(nullable=False),
        'name': object_fields.StringField(nullable=False),
        'groups': object_fields.ListOfDictOfNullableStringsField(),
        'description': object_fields.StringField(nullable=True),
    }

    def obj_make_compatible(self, primitive, target_version):
        super(DeviceProfile, self).obj_make_compatible(
            primitive, target_version)
        target_version = versionutils.convert_version_to_tuple(target_version)
        if target_version < (1, 1) and 'description' in primitive:
            del primitive['description']

    def _to_profile_json(self, obj_changes):
        if 'groups' in obj_changes:  # Convert to profile_json string
            d = {"groups": obj_changes['groups']}
            profile_json = jsonutils.dumps(d)
            obj_changes['profile_json'] = profile_json
            obj_changes.pop('groups', None)  # del 'groups'
        else:
            raise exception.DeviceProfileGroupsExpected()

    def create(self, context):
        """Create a Device Profile record in the DB."""
        # TODO() validate with a JSON schema
        if 'name' not in self:
            raise exception.ObjectActionError(action='create',
                                              reason='name is required')

        values = self.obj_get_changes()
        self._to_profile_json(values)

        db_devprof = self.dbapi.device_profile_create(context, values)
        self._from_db_object(self, db_devprof)

    @classmethod
    def get_by_id(cls, context, id):
        """Find a DB Device_profile and return an Obj Device_profile."""
        db_devprof = cls.dbapi.device_profile_get_by_id(context, id)
        obj_devprof = cls._from_db_object(cls(context), db_devprof)
        return obj_devprof

    @classmethod
    def get_by_uuid(cls, context, uuid):
        """Find a DB Device_profile and return an Obj Device_profile."""
        db_devprof = cls.dbapi.device_profile_get_by_uuid(context, uuid)
        obj_devprof = cls._from_db_object(cls(context), db_devprof)
        return obj_devprof

    @classmethod
    def get_by_name(cls, context, name):
        """Find a DB Device Profile and return an Obj Device Profile."""
        db_devprof = cls.dbapi.device_profile_get(context, name)
        obj_devprof = cls._from_db_object(cls(context), db_devprof)
        return obj_devprof

    @classmethod
    def list(cls, context):
        # TODO() add filters, limits, pagination, etc.
        """Return a list of Device Profile objects."""
        db_devprofs = cls.dbapi.device_profile_list(context)
        obj_dp_list = cls._from_db_object_list(db_devprofs, context)
        return obj_dp_list

    def save(self, context):
        """Update a Device Profile record in the DB."""
        updates = self.obj_get_changes()
        self._to_profile_json(updates)

        db_devprof = self.dbapi.device_profile_update(context,
                                                      self.name, updates)
        self._from_db_object(self, db_devprof)

    def destroy(self, context):
        """Delete a Device Profile from the DB."""
        self.dbapi.device_profile_delete(context, self.uuid)
        self.obj_reset_changes()

    @classmethod
    def _from_db_object(cls, obj, db_obj):
        """Converts a device_profile to a formal object.

        :param obj: An object of the class.
        :param db_obj: A DB model of the object
        :return: The object of the class with the database entity added
        """
        # Convert from profile_json to 'groups' ListOfDictOfStrings
        d = jsonutils.loads(db_obj['profile_json'])
        db_obj['groups'] = d['groups']
        obj = base.CyborgObject._from_db_object(obj, db_obj)
        return obj



# -*- coding: utf-8 -*-

# Copyright 2016-2017 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from sqlalchemy import Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base
Base = declarative_base()


# A common internal acclerator object for internal use.
class Accelerator(Base):
    __tablename__ = 'accelerators'
    accelerator_id = Column(String, primary_key=True)
    device_type = Column(String)
    remoteable = Column(Integer)
    vender_id = Column(String)
    product_id = Column(String)

    def __init__(self, **kwargs):
        self.accelerator_id = kwargs['accelerator_id']
        self.device_type = kwargs['device_type']
        self.remoteable = kwargs['remoteable']
        self.vendor_id = kwargs['vendor_id']
        self.product_id = kwargs['product_id']



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import fixtures
from oslo_config import cfg
from oslo_policy import opts as policy_opts

from cyborg.common import authorize_wsgi as cyborg_policy


CONF = cfg.CONF

policy_data = """
{

}
"""


class PolicyFixture(fixtures.Fixture):
    def setUp(self):
        super(PolicyFixture, self).setUp()
        self.policy_dir = self.useFixture(fixtures.TempDir())
        self.policy_file_name = os.path.join(self.policy_dir.path,
                                             'policy.yaml')
        with open(self.policy_file_name, 'w') as policy_file:
            policy_file.write(policy_data)
        policy_opts.set_defaults(CONF)
        CONF.set_override('policy_file', self.policy_file_name, 'oslo_policy')
        cyborg_policy._ENFORCER = None
        # NOTE(yumeng): Logging all the deprecation warning for every unit
        # test will overflow the log files and leads to error. Suppress
        # the deprecation warning for tests only.
        cyborg_policy.init_enforcer(suppress_deprecation_warnings=True)
        self.addCleanup(cyborg_policy.get_enforcer().clear)



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Database setup and migration commands."""

from oslo_config import cfg
from stevedore import driver


_IMPL = None


def get_backend():
    global _IMPL
    if not _IMPL:
        cfg.CONF.import_opt('backend', 'oslo_db.options', group='database')
        _IMPL = driver.DriverManager("cyborg.database.migration_backend",
                                     cfg.CONF.database.backend).driver
    return _IMPL


def upgrade(version=None):
    """Migrate the database to `version` or the most recent version."""
    return get_backend().upgrade(version)


def version():
    return get_backend().version()


def stamp(version):
    return get_backend().stamp(version)


def revision(message, autogenerate):
    return get_backend().revision(message, autogenerate)


def create_schema():
    return get_backend().create_schema()



# Copyright 2020 ZTE Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import glanceclient

from cyborg.common import exception
from cyborg.image import glance
from cyborg.tests.unit.db import base


class TestExceptionTranslations(base.DbTestCase):

    def test_client_notfound_converts_to_imagenotfound(self):
        in_exc = glanceclient.exc.NotFound('123')
        out_exc = glance._translate_image_exception('123', in_exc)
        self.assertIsInstance(out_exc, exception.ResourceNotFound)



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
import oslo_messaging as messaging
import uuid

from cyborg.common import exception
from cyborg.common import placement_client
from cyborg.conf import CONF
from cyborg.objects.attach_handle import AttachHandle
from cyborg.objects.attribute import Attribute
from cyborg.objects.control_path import ControlpathID
from cyborg.objects.deployable import Deployable
from cyborg.objects.device import Device
from cyborg.objects.driver_objects.driver_device import DriverDeployable
from cyborg.objects.driver_objects.driver_device import DriverDevice
from cyborg.objects.ext_arq import ExtARQ

LOG = logging.getLogger(__name__)


class ConductorManager(object):
    """Cyborg Conductor manager main class."""

    RPC_API_VERSION = '1.0'
    target = messaging.Target(version=RPC_API_VERSION)

    def __init__(self, topic, host=None):
        super(ConductorManager, self).__init__()
        self.topic = topic
        self.host = host or CONF.host
        self.placement_client = placement_client.PlacementClient()

    def periodic_tasks(self, context, raise_on_error=False):
        pass

    def device_profile_create(self, context, obj_devprof):
        """Signal to conductor service to create a device_profile.

        :param context: request context.
        :param obj_devprof: a created (but not saved) device_profile object.
        :returns: created device_profile object.
        """
        obj_devprof.create(context)
        return obj_devprof

    def device_profile_delete(self, context, obj_devprof):
        """Signal to conductor service to delete a device_profile.
        :param context: request context.
        :param obj_devprof: a device_profile object to delete.
        """
        obj_devprof.destroy(context)

    def arq_create(self, context, obj_extarq, devprof_id):
        """Signal to conductor service to create an accelerator requests.

        :param context: request context.
        :param obj_extarq: a created (but not saved) accelerator_requests
        object
        :param devprof_id: a device profile id
        :returns: saved accelerator_requests object.
        """
        obj_extarq.create(context, devprof_id)
        return obj_extarq

    def arq_delete_by_uuid(self, context, arqs):
        """Signal to conductor service to delete accelerator requests by
        ARQ UUIDs.

        :param context: request context.
        :param arqs: ARQ UUIDs joined with ','
        """
        arqlist = arqs.split(',')
        ExtARQ.delete_by_uuid(context, arqlist)

    def arq_delete_by_instance_uuid(self, context, instance):
        """Signal to conductor service to delete accelerator requests by
        instance UUID.

        :param context: request context.
        :param instance: UUID of instance whose ARQs need to be deleted
        """
        ExtARQ.delete_by_instance(context, instance)

    def arq_apply_patch(self, context, patch_list, valid_fields):
        """Signal to conductor service to apply patch accelerator requests.

        :param context: request context.
        :param patch_list: A map from ARQ UUIDs to their JSON patches
        :param valid_fields: Dict of valid fields
        """
        ExtARQ.apply_patch(context, patch_list, valid_fields)

    def report_data(self, context, hostname, driver_device_list):
        """Update the Cyborg DB in one hostname according to the
        discovered device list.
        :param context: request context.
        :param hostname: agent's hostname.
        :param driver_device_list: a list of driver_device object
        discovered by agent in the host.
        """
        # TODO(): Everytime get from the DB?
        # First retrieve the old_device_list from the DB.
        old_driver_device_list = DriverDevice.list(context, hostname)
        # TODO(wangzhh): Remove invalid driver_devices without controlpath_id.
        # Then diff two driver device list.
        self.drv_device_make_diff(context, hostname,
                                  old_driver_device_list, driver_device_list)

    def drv_device_make_diff(self, context, host, old_driver_device_list,
                             new_driver_device_list):
        """Compare new driver-side device object list with the old one in
        one host.
        """
        LOG.info("Start differing devices.")
        # TODO(): The placement report will be implemented here.
        # Use cpid.cpid_info to identify whether the device is the same.
        stub_cpid_list = [driver_dev_obj.controlpath_id.cpid_info for
                          driver_dev_obj in new_driver_device_list
                          if driver_dev_obj.stub]
        new_cpid_list = [driver_dev_obj.controlpath_id.cpid_info for
                         driver_dev_obj in new_driver_device_list]
        old_cpid_list = [driver_dev_obj.controlpath_id.cpid_info for
                         driver_dev_obj in old_driver_device_list]
        same = set(new_cpid_list) & set(old_cpid_list) - set(stub_cpid_list)
        added = set(new_cpid_list) - same - set(stub_cpid_list)
        deleted = set(old_cpid_list) - same - set(stub_cpid_list)
        host_rp = self._get_root_provider(context, host)
        # device is deleted.
        for d in deleted:
            old_driver_dev_obj = old_driver_device_list[old_cpid_list.index(d)]
            for driver_dep_obj in old_driver_dev_obj.deployable_list:
                rp_uuid = self.get_rp_uuid_from_obj(driver_dep_obj)
                self._delete_provider_and_sub_providers(context, rp_uuid)
            old_driver_dev_obj.destroy(context, host)
        # device is added
        for a in added:
            new_driver_dev_obj = new_driver_device_list[new_cpid_list.index(a)]
            try:
                new_driver_dev_obj.create(context, host)
            except Exception as exc:
                LOG.exception("Failed to add device %(device)s. "
                              "Reason: %(reason)s",
                              {'device': new_driver_dev_obj,
                               'reason': exc})
                new_driver_dev_obj.destroy(context, host)
            # TODO(All): If report device data to Placement raise exception,
            # we should revert driver device created in Cyborg and resources
            # created in Placement to reduce the risk of data inconsistency
            # here between Cyborg and Placement.
            cleanup_inconsistency_resources = False
            for driver_dep_obj in new_driver_dev_obj.deployable_list:
                try:
                    self.get_placement_needed_info_and_report(context,
                                                              driver_dep_obj,
                                                              host_rp)
                except Exception as exc:
                    LOG.info("Failed to add device %(device)s. "
                             "Reason: %(reason)s",
                             {'device': new_driver_dev_obj,
                              'reason': exc})
                    cleanup_inconsistency_resources = True
                    break
            if cleanup_inconsistency_resources:
                new_driver_dev_obj.destroy(context, host)
                for driver_dep_obj in new_driver_dev_obj.deployable_list:
                    rp_uuid = self.get_rp_uuid_from_obj(driver_dep_obj)
                    self._delete_provider_and_sub_providers(context, rp_uuid)
        for s in same:
            # get the driver_dev_obj, diff the driver_device layer
            new_driver_dev_obj = new_driver_device_list[new_cpid_list.index(s)]
            old_driver_dev_obj = old_driver_device_list[old_cpid_list.index(s)]
            # First, get dev_obj_list from hostname
            device_obj_list = Device.get_list_by_hostname(context, host)
            # Then, use controlpath_id.cpid_info to identiy one Device.
            cpid_info = new_driver_dev_obj.controlpath_id.cpid_info
            for dev_obj in device_obj_list:
                # get cpid_obj, could be empty or only one value.
                cpid_obj = ControlpathID.get_by_device_id_cpidinfo(
                    context, dev_obj.id, cpid_info)
                # find the one cpid_obj with cpid_info
                if cpid_obj is not None:
                    break

            changed_key = ['std_board_info', 'vendor', 'vendor_board_info',
                           'model', 'type']
            for c_k in changed_key:
                if getattr(new_driver_dev_obj, c_k) != getattr(
                        old_driver_dev_obj, c_k):
                    setattr(dev_obj, c_k, getattr(new_driver_dev_obj, c_k))
            dev_obj.save(context)
            # diff the internal layer: driver_deployable
            self.drv_deployable_make_diff(context, dev_obj.id, cpid_obj.id,
                                          old_driver_dev_obj.deployable_list,
                                          new_driver_dev_obj.deployable_list,
                                          host_rp)

    def drv_deployable_make_diff(self, context, device_id, cpid_id,
                                 old_driver_dep_list, new_driver_dep_list,
                                 host_rp):
        """Compare new driver-side deployable object list with the old one in
        one host.
        """
        # use name to identify whether the deployable is the same.
        LOG.info("Start differing deploybles.")
        new_name_list = [driver_dep_obj.name for driver_dep_obj in
                         new_driver_dep_list]
        old_name_list = [driver_dep_obj.name for driver_dep_obj in
                         old_driver_dep_list]
        same = set(new_name_list) & set(old_name_list)
        added = set(new_name_list) - same
        deleted = set(old_name_list) - same
        # name is deleted.
        for d in deleted:
            old_driver_dep_obj = old_driver_dep_list[old_name_list.index(d)]
            rp_uuid = self.get_rp_uuid_from_obj(old_driver_dep_obj)
            old_driver_dep_obj.destroy(context, device_id)
            self._delete_provider_and_sub_providers(context, rp_uuid)
        # name is added.
        for a in added:
            new_driver_dep_obj = new_driver_dep_list[new_name_list.index(a)]
            new_driver_dep_obj.create(context, device_id, cpid_id)
            try:
                self.get_placement_needed_info_and_report(context,
                                                          new_driver_dep_obj,
                                                          host_rp)
            except Exception as exc:
                LOG.info("Failed to add deployable %(deployable)s. "
                         "Reason: %(reason)s",
                         {'deployable': new_driver_dep_obj,
                          'reason': exc})
                new_driver_dep_obj.destroy(context, device_id)
                rp_uuid = self.get_rp_uuid_from_obj(new_driver_dep_obj)
                # TODO(All): If report deployable data to Placement raise
                # exception, we should revert driver deployable created in
                # Cyborg and resources created in Placement to reduce the risk
                # of data inconsistency here between Cyborg and Placement.
                self._delete_provider_and_sub_providers(context, rp_uuid)
        for s in same:
            # get the driver_dep_obj, diff the driver_dep layer
            new_driver_dep_obj = new_driver_dep_list[new_name_list.index(s)]
            old_driver_dep_obj = old_driver_dep_list[old_name_list.index(s)]
            # get dep_obj, it won't be None because it stored before.
            dep_obj = Deployable.get_by_name_deviceid(context, s, device_id)
            # update the driver_dep num_accelerators field
            if dep_obj.num_accelerators != new_driver_dep_obj.num_accelerators:
                dep_obj.num_accelerators = new_driver_dep_obj.num_accelerators
                dep_obj.save(context)
                rp_uuid = self.get_rp_uuid_from_obj(new_driver_dep_obj)
                attrs = new_driver_dep_obj.attribute_list
                resource_class = [i.value for i in attrs if i.key == 'rc'][0]
                inv_data = _gen_resource_inventory(
                    resource_class, dep_obj.num_accelerators)
                self.placement_client.update_inventory(rp_uuid, inv_data)
            # diff the internal layer: driver_attribute_list
            new_attribute_list = []
            if hasattr(new_driver_dep_obj, 'attribute_list'):
                new_attribute_list = new_driver_dep_obj.attribute_list
            self.drv_attr_make_diff(context, dep_obj.id,
                                    old_driver_dep_obj.attribute_list,
                                    new_attribute_list)
            # diff the internal layer: driver_attach_hanle_list
            self.drv_ah_make_diff(context, dep_obj.id, cpid_id,
                                  old_driver_dep_obj.attach_handle_list,
                                  new_driver_dep_obj.attach_handle_list)

    def drv_attr_make_diff(self, context, dep_id, old_driver_attr_list,
                           new_driver_attr_list):
        """Diff new driver-side Attribute Object lists with the old one."""
        LOG.info("Start differing attributes.")
        dep_obj = Deployable.get_by_id(context, dep_id)
        driver_dep = DriverDeployable.get_by_name(context, dep_obj.name)
        rp_uuid = self.get_rp_uuid_from_obj(driver_dep)
        new_key_list = [driver_attr_obj.key for driver_attr_obj in
                        new_driver_attr_list]
        old_key_list = [driver_attr_obj.key for driver_attr_obj in
                        old_driver_attr_list]
        same = set(new_key_list) & set(old_key_list)
        # key is deleted.
        deleted = set(old_key_list) - same
        for d in deleted:
            old_driver_attr_obj = old_driver_attr_list[old_key_list.index(d)]
            self.placement_client.delete_trait_by_name(
                context, rp_uuid, old_driver_attr_obj.value)
            old_driver_attr_obj.delete_by_key(context, dep_id, d)
        # key is added.
        added = set(new_key_list) - same
        for a in added:
            new_driver_attr_obj = new_driver_attr_list[new_key_list.index(a)]
            new_driver_attr_obj.create(context, dep_id)
            self.placement_client.add_traits_to_rp(
                rp_uuid, [new_driver_attr_obj.value])
        # key is same, diff the value.
        for s in same:
            # value is not same, update
            new_driver_attr_obj = new_driver_attr_list[new_key_list.index(s)]
            old_driver_attr_obj = old_driver_attr_list[old_key_list.index(s)]
            if new_driver_attr_obj.value != old_driver_attr_obj.value:
                attr_obj = Attribute.get_by_dep_key(context, dep_id, s)
                attr_obj.value = new_driver_attr_obj.value
                attr_obj.save(context)
                # Update traits here.
                if new_driver_attr_obj.key.startswith("trait"):
                    self.placement_client.delete_trait_by_name(
                        context, rp_uuid, old_driver_attr_obj.value)
                    self.placement_client.add_traits_to_rp(
                        rp_uuid, [new_driver_attr_obj.value])
                # Update resource classes here.
                if new_driver_attr_obj.key.startswith("rc"):
                    self.placement_client.ensure_resource_classes(
                        context, [new_driver_attr_obj.value])
                    inv_data = _gen_resource_inventory(
                        new_driver_attr_obj.value, dep_obj.num_accelerators)
                    self.placement_client.update_inventory(rp_uuid, inv_data)
                    self.placement_client.delete_rc_by_name(
                        context, old_driver_attr_obj.value)

    @classmethod
    def drv_ah_make_diff(cls, context, dep_id, cpid_id, old_driver_ah_list,
                         new_driver_ah_list):
        """Diff new driver-side AttachHandle Object lists with the old one."""
        LOG.info("Start differing attach_handles.")
        new_info_list = [driver_ah_obj.attach_info for driver_ah_obj in
                         new_driver_ah_list]
        old_info_list = [driver_ah_obj.attach_info for driver_ah_obj in
                         old_driver_ah_list]
        same = set(new_info_list) & set(old_info_list)
        LOG.info('new info list %s', new_info_list)
        LOG.info('old info list %s', old_info_list)
        # attach_info is deleted.
        deleted = set(old_info_list) - same
        for d in deleted:
            old_driver_ah_obj = old_driver_ah_list[old_info_list.index(d)]
            old_driver_ah_obj.destroy(context, dep_id)
        # attach_info is added.
        added = set(new_info_list) - same
        for a in added:
            new_driver_ah_obj = new_driver_ah_list[new_info_list.index(a)]
            new_driver_ah_obj.create(context, dep_id, cpid_id)
        # attach-info is same
        for s in same:
            # get attach_handle obj
            new_driver_ah_obj = new_driver_ah_list[new_info_list.index(s)]
            old_driver_ah_obj = old_driver_ah_list[old_info_list.index(s)]
            changed_key = ['attach_type']
            ah_obj = AttachHandle.get_ah_by_depid_attachinfo(context,
                                                             dep_id, s)
            for c_k in changed_key:
                if getattr(new_driver_ah_obj, c_k) != getattr(
                        old_driver_ah_obj, c_k):
                    setattr(ah_obj, c_k, getattr(new_driver_ah_obj, c_k))
            ah_obj.save(context)

    def _get_root_provider(self, context, hostname):
        try:
            provider = self.placement_client.get(
                "resource_providers?name=" + hostname).json()
            pr_uuid = provider["resource_providers"][0]["uuid"]
            return pr_uuid
        except (IndexError, KeyError):
            raise exception.PlacementResourceProviderNotFound(
                resource_provider=hostname)

    def _get_sub_provider(self, context, parent, name):
        old_sub_pr_uuid = str(uuid.uuid3(uuid.NAMESPACE_DNS,
                                         str(name)))
        new_sub_pr_uuid = self.placement_client.ensure_resource_provider(
            context, old_sub_pr_uuid,
            name=name, parent_provider_uuid=parent)
        if old_sub_pr_uuid == new_sub_pr_uuid:
            return new_sub_pr_uuid
        else:
            raise exception.Conflict()

    def provider_report(self, context, name, resource_class, traits, total,
                        parent):
        self.placement_client.ensure_resource_classes(
            context, [resource_class])
        sub_pr_uuid = self._get_sub_provider(
            context, parent, name)
        result = _gen_resource_inventory(resource_class, total)
        self.placement_client.update_inventory(sub_pr_uuid, result)
        # traits = ["CUSTOM_FPGA_INTEL", "CUSTOM_FPGA_INTEL_ARRIA10",
        #           "CUSTOM_FPGA_INTEL_REGION_UUID",
        #           "CUSTOM_FPGA_FUNCTION_ID_INTEL_UUID",
        #           "CUSTOM_PROGRAMMABLE",
        #           "CUSTOM_FPGA_NETWORK"]
        self.placement_client.add_traits_to_rp(sub_pr_uuid, traits)
        return sub_pr_uuid

    def get_placement_needed_info_and_report(self, context, obj,
                                             parent_uuid=None):
        pr_name = obj.name
        attrs = obj.attribute_list
        resource_class = [i.value for i in attrs if i.key == 'rc'][0]
        traits = [i.value for i in attrs
                  if str(i.key).startswith("trait")]
        total = obj.num_accelerators
        rp_uuid = self.provider_report(context, pr_name, resource_class,
                                       traits, total, parent_uuid)
        dep_obj = Deployable.get_by_name(context, pr_name)
        dep_obj["rp_uuid"] = rp_uuid
        dep_obj.save(context)

    def get_rp_uuid_from_obj(self, obj):
        return str(uuid.uuid3(uuid.NAMESPACE_DNS, str(obj.name)))

    def _delete_provider_and_sub_providers(self, context, rp_uuid):
        rp_in_tree = self.placement_client.get_providers_in_tree(context,
                                                                 rp_uuid)
        for rp in rp_in_tree[::-1]:
            if rp["parent_provider_uuid"] == rp_uuid or rp["uuid"] == rp_uuid:
                self.placement_client.delete_provider(rp["uuid"])
                LOG.info("Sucessfully delete resource provider %(rp_uuid)s",
                         {"rp_uuid": rp["uuid"]})
                if rp["uuid"] == rp_uuid:
                    break


def _gen_resource_inventory(resource_class, total):
    return {
        resource_class: {
            'total': total,
            'max_unit': total,
        },
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import psutil
import re
import subprocess

from oslo_concurrency import processutils
from oslo_log import log as logging

import cyborg.privsep

LOG = logging.getLogger(__name__)


class PySPDK(object):

    def __init__(self, pname):
        super(PySPDK, self).__init__()
        self.pid = None
        self.pname = pname

    @cyborg.privsep.sys_admin_pctxt.entrypoint
    def start_server(self, spdk_dir, server_name):
        if not self.is_alive():
            self.init_hugepages(spdk_dir)
            server_dir = os.path.join(spdk_dir, 'app/')
            file_dir = self._search_file(server_dir, server_name)
            LOG.info(file_dir)
            os.chdir(file_dir)
            cmd = ['bash', server_name]
            out, err = processutils.execute(*cmd)
            return out

    @cyborg.privsep.sys_admin_pctxt.entrypoint
    def init_hugepages(self, spdk_dir):
        huge_dir = os.path.join(spdk_dir, 'scripts/')
        file_dir = self._search_file(huge_dir, 'setup.sh')
        LOG.info(file_dir)
        os.chdir(file_dir)
        cmd = ['bash', 'setup.sh']
        out, err = processutils.execute(*cmd)
        return out

    @staticmethod
    def _search_file(spdk_dir, file_name):
        for dirpath, dirnames, filenames in os.walk(spdk_dir):
            for filename in filenames:
                if filename == file_name:
                    return dirpath

    def _get_process_id(self):
        for proc in psutil.process_iter():
            try:
                pinfo = proc.as_dict(attrs=['pid', 'cmdline'])
                if re.search(self.pname, str(pinfo.get('cmdline'))):
                    self.pid = pinfo.get('pid')
                    return self.pid
            except psutil.NoSuchProcess:
                LOG.info("NoSuchProcess:%(pname)s", {"pname": self.pname})
        LOG.info("NoSuchProcess:%(pname)s", {"pname": self.pname})
        return self.pid

    def is_alive(self):
        self.pid = self._get_process_id()
        if self.pid:
            p = psutil.Process(self.pid)
            if p.is_running():
                return True
        return False

    @staticmethod
    def exec_rpc(method, server='127.0.0.1', port=5260, sub_args=None):
        exec_cmd = ["./rpc.py", "-s", "-p"]
        exec_cmd.insert(2, server)
        exec_cmd.insert(4, str(port))
        exec_cmd.insert(5, method)
        if sub_args is None:
            sub_args = []
        exec_cmd.extend(sub_args)
        p = subprocess.Popen(
            exec_cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        out, err = p.communicate()
        return out
